from __future__ import annotations

import abc
from dataclasses import dataclass
from typing import Callable

import numpy as np
from numpy.typing import ArrayLike, NDArray

from boss.utils.arrays import shape_consistent, shape_consistent_XY


@dataclass
class UserFuncOutput:
    """Stores the output from the user function(s)."""

    X: NDArray[np.float64]
    Y: NDArray[np.float64]


class UserFunc(abc.ABC):
    """Abstract base class for BOSS user function wrapper classes."""

    def __init__(self, func: Callable, dim: int) -> None:
        """
        Initializes a new wrapped user function.

        Parameters
        ----------
        func : Callable
            The user function to wrap.
        dim : int
            The dimension of the input space.
        """
        self.func = func
        self.dim = dim

    @abc.abstractmethod
    def evaluate(self, X_in: ArrayLike) -> UserFuncOutput:
        """Evaluates the user function at the given inputs."""

    def __call__(self, X_in: ArrayLike) -> UserFuncOutput:
        """
        Alias for UserFunc.evaluate.
        """
        return self.evaluate(X_in)


class SingleTask(UserFunc):
    """Wrapper class for BOSS user functions.

    Checks for extra user-defined data (e.g, in the case of symmetry points).
    """

    def evaluate(self, X_in: ArrayLike) -> UserFuncOutput:
        """
        Evaluates the user function (possibly with gradients) at the given inputs.

        Note that more inputs maybe be added by the user function, e.g.,
        in the case of symmetry points.

        Parameters
        ----------
        X_in : np.ndarray with shape (n_inputs, dim)
            The inputs for which to evaluate the user function.

        Returns
        -------
        UserFuncOutput
            Dataclass with attributes:
                X : np.ndarray with shape (n_inputs, dim)
                    The inputs used to evaluate the user function.
                Y : np.ndarray with shape (n_inputs, 1) or (n_inputs, 1 + dim)
                    Corresponding array of user function values, if there are
                    gradients they are stored in the last columns.
        """
        X_in = np.atleast_2d(X_in)
        Y = []
        X_out = []
        # Loop over batch points
        for x_in in X_in:
            out = self.func(np.atleast_2d(x_in))  # 2d input for backwards compat.
            # check if userfn returns extra data (e.g., symmetry points)
            if isinstance(out, tuple):
                x_out, y = shape_consistent_XY(*out, x_dim=self.dim)
                X_out.append(x_out)
                Y.append(y)
            else:
                y = shape_consistent(out, dim=1)
                Y.append(y)

        if len(X_out) == 0:
            X_out = X_in
        else:
            X_out = np.concatenate(X_out)

        Y = np.concatenate(Y)
        return UserFuncOutput(X_out, Y)


class SingleTaskGradient(UserFunc):
    """Wrapper class for BOSS user functions with gradient observations.

    Automatically  extra user-defined data (e.g, in the case of symmetry points).
    """

    def evaluate(self, X_in: ArrayLike) -> UserFuncOutput:
        """
        Evaluates the user function (possibly with gradients) at the given inputs.

        Note that more inputs maybe be added by the user function, e.g.,
        in the case of symmetry points.

        Parameters
        ----------
        X_in : np.ndarray with shape (n_inputs, dim)
            The inputs for which to evaluate the user function.

        Returns
        -------
        UserFuncOutput
            Dataclass with attributes:
                X : np.ndarray with shape (n_inputs, dim)
                    The inputs used to evaluate the user function.
                Y : np.ndarray with shape (n_inputs, 1) or (n_inputs, 1 + dim)
                    Corresponding array of user function values, if there are
                    gradients they are stored in the last columns.
        """
        X_in = np.atleast_2d(X_in)
        Y = []
        X_out = []
        # Loop over batch points
        for x_in in X_in:
            out = self.func(np.atleast_2d(x_in))
            # check if userfn returns extra data (e.g., symmetry points)
            if len(out) == 3:
                x_out, y, dy = out
                x_out, y = shape_consistent_XY(x_out, y, x_dim=self.dim)
                dy, y = shape_consistent_XY(dy, y, x_dim=self.dim)
                X_out.append(x_out)
            elif len(out) == 2:
                y, dy = out
                dy, y = shape_consistent_XY(dy, y, x_dim=self.dim)
            else:
                raise ValueError("Incorrect number of ouputs from user function.")

            Y.append(np.concatenate((y, dy), axis=1))

        if len(X_out) == 0:
            X_out = X_in
        else:
            X_out = np.concatenate(X_out)

        Y = np.concatenate(Y)
        return UserFuncOutput(X_out, Y)


class MultiTask(UserFunc):
    """Wrapper class for multiple BOSS user functions."""

    def __init__(self, func_list: list[Callable], dim: int) -> None:
        self.func_list = func_list
        self.dim = dim

    def __len__(self) -> int:
        """Returns the length of the user function list."""
        return len(self.func_list)

    def evaluate(self, X_in: ArrayLike) -> UserFuncOutput:
        """
        Evaluates the user function at the given inputs.

        Parameters
        ----------
        X_in : np.ndarray with shape (n_inputs, dim + 1)
            Input data and corresponding task indices for the user functions
            to be evaluated.

        Returns
        -------
        UserFuncOutput
            Dataclass with attributes:
                X : np.ndarray with shape (n_inputs, dim)
                    The inputs used to evaluate the user function.
                Y : np.ndarray with shape (n_inputs, 1) or (n_inputs, 1 + dim)
                    Corresponding array of user function values, if there are
                    gradients they are stored in the last columns.
        """
        # Separate x-data and indices
        X_in = np.atleast_2d(X_in)
        X_in, indices = X_in[:, :-1], X_in[:, -1]
        indices = indices.astype(int)

        Y = []
        X_out = []
        indices_out = []
        # Loop over batch points
        for x_in, index in zip(X_in, indices):
            out = self.func_list[index](np.atleast_2d(x_in))
            # check if userfn returns extra data (e.g., symmetry points)
            if isinstance(out, tuple):
                x_out, y = out
                x_out, y = shape_consistent_XY(x_out, y, x_dim=self.dim)
                index_out = index * np.ones(x_out.shape[0], dtype=float)
                indices_out.append(index_out)
                X_out.append(x_out)
                Y.append(y)
            else:
                y = shape_consistent(out, dim=1)
                Y.append(y)

        if len(X_out) == 0:
            X_out = X_in
            indices_out = indices[:, None]
        else:
            indices_out = np.concatenate(indices_out)[:, None]

        # put X data and indices back together
        X_out = np.concatenate((X_out, indices_out), axis=1)
        Y = np.concatenate(Y)
        return UserFuncOutput(X_out, Y)



from scipy.interpolate import griddata
from itertools import product
import numpy as np
import os
import shutil
from boss.pp.mesh import Mesh


class PFMain:
    def __init__(
        self,
        n_tasks=2,
        pf_mesh=30,
        bounds=np.array([[0, 1.0], [0, 1.0]]),
        pf_optimal_solution=True,
        pf_correlation_maps=True,
        pf_optimal_weights=None,
        pf_optimal_reference=None,
        pf_optimal_num_sol=1,
        pf_optimal_order=2,
        models=None,
    ):
        """
        Calculates the Pareto front and optimal solutions for multi-objective optimization.

        - Overview:
            - This class finds the Pareto front and corresponding solutions for problems with any number of input dimensions and objectives.


        Parameters
        ----------
        n_tasks : int
            The number of objectives to optimize.

        pf_mesh : int
            The size of the discretized data for each model.

        bounds : ndarray
            The bounds for each model.

        models : boss result function
            The BOSS GP models.

        Returns
        -------
        None
            This function does not return anything directly, but performs calculations
            to determine the Pareto front and the Pareto optimal solutions.
        """

        self.n_tasks = n_tasks
        self.pf_mesh = pf_mesh
        self.bounds = bounds
        self.n_dim = np.shape(self.bounds)[0]
        self.models = models
        self.n_mesh = (self.pf_mesh,) * self.n_dim
        self.mesher = Mesh(self.bounds, grid_pts=self.pf_mesh)
        self.X_mg = self.mesher.grid
        self.pf_optimal_solution = pf_optimal_solution
        self.pf_optimal_weights = pf_optimal_weights
        self.pf_optimal_reference = pf_optimal_reference
        self.pf_optimal_num_sol = pf_optimal_num_sol
        self.pf_optimal_order = pf_optimal_order
        self.pf_correlation_maps = pf_correlation_maps

    def get_data_on_grid(self, path, bounds, x_pts):
        """
        Retrieve data on a grid.

        This subroutine is a needed function when meshes don't match from different tasks.

        Parameters
        ----------
        path : str
            Path to the models.

        bounds : ndarray
            The bounds for each model.

        x_pts : ndarray
            All input data points.

        Returns
        -------
        ndarray
            Numpy array with models evaluated on a uniform mesh grid
        """

        data = np.loadtxt(path)
        x_inp = tuple([data[:, i] for i in range(self.n_dim)])
        y_inp = data[:, self.n_dim]
        x_pts_int = tuple([x_pts[i] for i in range(self.n_dim)])
        gd = griddata(x_inp, y_inp, x_pts_int, method="nearest")
        return gd

    def x_mesh(self, bounds, n_mesh):
        """
         Generates a linear space between the edges of the bounds for each input dimension.

         Parameters
         ----------

         bounds : ndarray
             The bounds for each model.

         n_mesh : ndarray
             An array of shape pf_mesh * model dimension

        Returns
        -------
        ndarray
             Numpy array with shape (n_dimension, mesh size)
        """
        x = [
            np.linspace(self.bounds[i, 0], self.bounds[i, 1], n_mesh[i])
            for i in range(self.n_dim)
        ]
        return x

    def position_coordinates(self, coordinate_grid, n_dim, values):
        """
        Arranges input data in n-dimensional space to correspond to model predictions for each objective.

        Parameters
        ----------
        coordinate_grid : ndarray
            Input data on the defined grid.

        n_dim : int
            Input dimensionality of each model.

        values : ndarray
            Corresponding prediction for each objective.

        Returns
        -------
        ndarray

            Numpy array with shape (model dimension, )
        """
        positions_tmp = coordinate_grid
        for i in range(self.n_dim):
            positions_tmp = positions_tmp.take(indices=values[i], axis=1)
        return positions_tmp

    def evaluate_mesh(self, f, X_mg):
        """
        Evaluates model predictions over a mesh.

        Parameters
        ----------
        f : function
            A function that returns an numpy array of the evaluated matrix on the uniform meshgrid.

        X_mg : ndarray
            See mesher class for more information

        Returns
        -------
        ndarray

            Numpy array with shape (mesh size, mesh size ......, n_dimensions times )
        """
        x_ind = []

        for i in range(self.n_dim):
            if self.n_dim > 1:
                x_ind.append(np.take(self.X_mg[i], 0, axis=i))
            else:
                x_ind.append(self.X_mg[i])
        nx_ind = []
        for i in range(self.n_dim):
            nx_ind.append(len(x_ind[i]))
        F = []
        bounds_list = [[] for i in range(self.n_dim)]
        for i in range(self.n_dim):
            bounds_list[i] = nx_ind[i]
        bounds_list = np.reshape(bounds_list, (self.n_dim, 1))
        for value in product(*(range(*bl) for bl in bounds_list)):
            values = [value[xx] for xx in range(self.n_dim)]
            coordinate_grid = np.array([self.X_mg[i] for i in range(self.n_dim)])
            positions = self.position_coordinates(coordinate_grid, self.n_dim, values)
            F.append(f(positions))
        F = np.array(F).reshape(tuple(nx_ind))
        return F

    def squeeze_mesh(self, n_size, data_pts):
        """
        Flattens an N-dimensional array.

        Parameters
        ----------
        n_size : int
            Input dimensionality of each model.

        data_pts : ndarray
            Input data on the defined grid.

        Returns
        -------
        ndarray

            Numpy array with shape (mesh size^n_dimensions, n_dimensions )
        """

        Y_squeezed_tmp = [data_pts[i].flatten()[:, None] for i in range(n_size)]
        Y_squeezed = np.hstack(tuple(Y_squeezed_tmp))
        return Y_squeezed

    def is_pareto(self, Y):
        """
        Checks if a point in an N-dimensional tensor is Pareto or not.

        Parameters
        ----------
        Y : ndarray
            Corresponding prediction for each objective.

        Returns
        -------
        ndarray
            Boolean array that corresponds to the index of the meshgrid which is a Pareto front/optimal solution.
        """
        pareto_mask = np.ones(Y.shape[0], dtype=bool)
        for i, y in enumerate(Y):
            if np.any(np.all(y < Y, axis=1)):
                pareto_mask[i] = False
        return pareto_mask

    def is_pareto_efficient(self, Y, return_mask=True, maximize=True):
        """
        Finds the Pareto-efficient points in a given set of points.

        Parameters
        ----------
        Y : array-like of shape (n_points, n_costs)
            The array of points, where each row represents a point and each column
            represents a cost or objective function value.
        return_mask : bool, optional
            If True, returns a boolean mask indicating which points are Pareto-efficient.
            If False, returns the indices of Pareto-efficient points. Default is True.
        maximize : bool, optional
            If True, treats the problem as a maximization problem (higher values are better).
            If False, treats the problem as a minimization problem (lower values are better).
            Default is False.

        Returns
        -------
        pareto_mask or pareto_indices : array-like
            If `return_mask` is True, returns a boolean array of shape (n_points,) indicating
            which points are Pareto-efficient.
            If `return_mask` is False, returns an integer array of indices of Pareto-efficient points.
        """
        if maximize:
            comparison = np.greater
        else:
            comparison = np.less

        is_efficient = np.arange(Y.shape[0])
        n_points = Y.shape[0]
        next_point_index = 0  # Next index in the is_efficient array to search for
        while next_point_index < len(Y):
            nondominated_point_mask = np.any(comparison(Y, Y[next_point_index]), axis=1)
            nondominated_point_mask[next_point_index] = True
            is_efficient = is_efficient[
                nondominated_point_mask
            ]  # Remove dominated points
            Y = Y[nondominated_point_mask]
            next_point_index = np.sum(nondominated_point_mask[:next_point_index]) + 1
        if return_mask:
            is_efficient_mask = np.zeros(n_points, dtype=bool)
            is_efficient_mask[is_efficient] = True
            return is_efficient_mask
        else:
            return is_efficient

    def correlation_maps(self):
        import matplotlib.pyplot as plt

        arrays = [
            np.loadtxt(f"models/grid_model_obj_{i}.dat")[:, -1]
            for i in range(self.n_tasks)
        ]

        array_names = [f"Y{i}" for i in range(self.n_tasks)]

        correlation_matrix = np.corrcoef(arrays)

        plt.imshow(correlation_matrix, cmap="coolwarm", interpolation="nearest")

        correlation_matrix = np.corrcoef(arrays)
        plt.xticks(ticks=np.arange(len(array_names)), labels=array_names)
        plt.yticks(ticks=np.arange(len(array_names)), labels=array_names)

        plt.colorbar(label="Correlation Coefficient")

        for i in range(len(array_names)):
            for j in range(len(array_names)):
                plt.text(
                    j,
                    i,
                    f"{correlation_matrix[i, j]:.2f}",
                    ha="center",
                    va="center",
                    color="black",
                )

        plt.title("Pearson Correlation Matrix Heatmap")
        plt.savefig(
            "postprocessing/pareto/pf_graph_models/correlation_matrix_heatmap.png"
        )

    def calc_pareto(self, functions, bounds, n_mesh):
        """
        Calculates the Pareto optimal solutions for any input dimensions and any number of objectives.

        Parameters
        ----------
        functions : ndarray
            Models of the various objectives.

        bounds : ndarray
            The bounds for each model.

        n_mesh : ndarray
            Output of the function n_mesh_generator.

        Returns
        -------
        pareto_solutions : ndarray
            Pareto optimal solutions (shape: number of dimensions, number of solutions).

        pareto_front : ndarray
            Pareto front (number of objectives, number of solutions).

        feasible_points : ndarray
            Feasible points (meshsize^number of dimensions, number of objectives).
        """
        func = [functions[i] for i in range(self.n_tasks)]
        all_Ys = [self.evaluate_mesh(func[i], self.X_mg) for i in range(self.n_tasks)]
        Xdata_pts = [self.X_mg[i] for i in range(self.n_dim)]
        X = self.squeeze_mesh(self.n_dim, Xdata_pts)
        Ydata_pts = [all_Ys[i] for i in range(self.n_tasks)]
        Y_feasible = self.squeeze_mesh(self.n_tasks, Ydata_pts)
        pareto_mask = self.is_pareto_efficient(Y_feasible)
        Y_pareto = Y_feasible[pareto_mask]
        X_pareto = X[pareto_mask]
        return X_pareto, Y_pareto, Y_feasible

    def all_functions(self, path):
        """
        Creates functions in the format needed for Pareto code for different mesh sizes.

        Parameters
        ----------
        path : str
            Paths to the output to n-dimensional BOSS models.

        Returns
        -------
        ndarray
            A function that returns an numpy array of the evaluated matrix on the uniform meshgrid.
        """

        def func(X):
            return self.get_data_on_grid(path, self.bounds, X)

        return func

    def boss_functions(self):
        """
        Calculates the prediction of the model on an n-dimensional grid.

        Returns
        -------
        function
            A function that returns an ndarray of the BOSS predictions model.
        """

        def f(x):
            return self.models.predict(np.atleast_2d(x))[0]

        return f

    def decision_maker(self):
        """
        A decision maker using the reference point EMO (Evolutionary multi-objective optimization) algorithm for a set of Pareto optimal solutions.

        Description
        -----------
        The normalized Euclidean distance (dij) between each solution of the best non-dominance level and the reference point is calculated.

        Returns
        -------
        ndarray
            Arrays corresponding to the decision maker
        """

        y_par = np.loadtxt("postprocessing/pareto/pf_data_models/y_pareto_output.dat")
        x_par = np.loadtxt("postprocessing/pareto/pf_data_models/x_pareto_output.dat")

        if self.n_dim == 1:
            n_dim = 1
        else:
            n_dim = x_par.shape[1]
        n_tasks = y_par.shape[1]

        fij_max = np.max(y_par, axis=0)
        fij_min = np.min(y_par, axis=0)

        if self.pf_optimal_weights is None:
            self.pf_optimal_weights = np.ones(n_tasks)

        if self.pf_optimal_reference is None:
            self.pf_optimal_reference = fij_max

        normalized_diff = (y_par - self.pf_optimal_reference) / (fij_max - fij_min)
        weighted_diff = self.pf_optimal_weights * normalized_diff
        dij_emo = (np.sum(weighted_diff**self.pf_optimal_order, axis=1)) ** (
            1 / self.pf_optimal_order
        )

        # Identify the positions of the best solutions
        pos_emo = np.argsort(dij_emo)[: self.pf_optimal_num_sol]
        pos_emo_min = pos_emo[0]

        print(
            "reference point, optimal-weights, pf-order, num_sol: ",
            self.pf_optimal_reference,
            self.pf_optimal_weights,
            self.pf_optimal_order,
            self.pf_optimal_num_sol,
        )

        # Save the selected solutions to files
        np.savetxt(
            "postprocessing/pareto/pf_data_models/x_reference.dat",
            x_par[pos_emo],
            fmt="%s",
        )
        np.savetxt(
            "postprocessing/pareto/pf_data_models/y_reference.dat",
            y_par[pos_emo],
            fmt="%s",
        )

        return np.atleast_2d(x_par[pos_emo_min]), np.atleast_2d(y_par[pos_emo_min])

    def write_output(self, file_name, Y, n_size):
        """
        Creates outputs in n columns for plotting.

        Parameters
        ----------
        file_name : str
            Output file name.

        Y : ndarray
            Pareto and feasible points.

        n_size : int
            Input dimensionality of the models.
        """
        with open(file_name, "w") as p_file:
            for i in range(len(Y)):
                for count in range(n_size):
                    if count == n_size - 1:
                        p_file.write("%s\n" % Y[i][count])
                    else:
                        p_file.write("%s " % Y[i][count])

    def generate_grid_models(self):
        """
        Generates an N-dimensional discretized Gaussian Process model if 'models' folder isn't created.
        """
        if self.models is not None:
            # boss_gp_model = self.models
            # X_mg = self.build_mesh(self.bounds, self.n_mesh)
            X_mg_ind = [self.X_mg[i].ravel() for i in range(self.n_dim)]
            if self.n_tasks > 1:
                Y = [
                    self.models.model.predict(np.column_stack((X_mg_ind)), index=i)[0]
                    for i in range(self.n_tasks)
                ]
            else:
                Y = [
                    self.models.model.predict(np.column_stack((X_mg_ind)))[0]
                    for i in range(self.n_tasks)
                ]
            xp_ind = [
                self.x_mesh(self.bounds, self.n_mesh)[i] for i in range(self.n_dim)
            ]
            boundaries_list = [self.n_mesh[j] for j in range(self.n_dim)]
            boundaries_list = np.reshape(boundaries_list, (self.n_dim, 1))
            # Check if the folder exists
            if os.path.exists("models"):
                # Delete the existing folder
                print("warning: overwriting folder 'models'")
                shutil.rmtree("models")
            os.mkdir("models")
            for nt in range(self.n_tasks):
                with open("models/grid_model_obj_%s.dat" % nt, "w") as file_tmp:
                    count_tmp = 0
                    for values in product(*(range(*b) for b in boundaries_list)):
                        for ind in range(self.n_dim):
                            file_tmp.write(
                                "%s " % (xp_ind[ind][values[self.n_dim - 1 - ind]])
                            )
                        file_tmp.write("%s\n" % (Y[nt][count_tmp][0]))
                        count_tmp += 1

    def get_pareto(self):
        """
        Starts the Pareto calculations and returns the Pareto fronts, optimal solutions,
        and feasible points from the function calc_pareto.

        Returns
        -------
        pareto_solutions : ndarray
            Pareto optimal solutions (shape: number of dimensions, number of solutions).

        pareto_front : ndarray
            Pareto front (number of objectives, number of solutions).

        feasible_points : ndarray
            Feasible points (meshsize^number of dimensions, number of objectives).
        """
        # self.input_path_generator()
        paths = []
        for i in range(self.n_tasks):
            paths.append("models/grid_model_obj_%s.dat" % i)

        self.generate_grid_models()
        functions = [self.all_functions(paths[i]) for i in range(self.n_tasks)]
        X_pareto, Y_pareto, Y_feasible = self.calc_pareto(
            functions, self.bounds, self.n_mesh
        )
        return X_pareto, Y_pareto, Y_feasible

    def run(self):
        """
        Runs the paretofront postprocessing routine.
        """
        X_pareto, Y_pareto, Y_feasible = self.get_pareto()

        # Check if the folder exists
        if os.path.exists("postprocessing/pareto"):
            # Delete the existing folder
            print("warning: overwriting folder 'postprocessing/pareto'")
            shutil.rmtree("postprocessing/pareto")
        if os.path.exists("postprocessing/"):
            os.mkdir("postprocessing/pareto")
        else:
            # print("postprocessing file does not exist, making one")
            os.mkdir("postprocessing")
            os.mkdir("postprocessing/pareto")
        os.mkdir("postprocessing/pareto/pf_data_models")
        os.mkdir("postprocessing/pareto/pf_graph_models")
        self.write_output(
            "postprocessing/pareto/pf_data_models/y_pareto_output.dat",
            Y_pareto,
            self.n_tasks,
        )
        self.write_output(
            "postprocessing/pareto/pf_data_models/x_pareto_output.dat",
            X_pareto,
            self.n_dim,
        )
        self.write_output(
            "postprocessing/pareto/pf_data_models/y_feasible_output.dat",
            Y_feasible,
            self.n_tasks,
        )

        # decision maker
        if self.pf_optimal_solution:
            self.decision_maker()

        if self.pf_correlation_maps:
            self.correlation_maps()

    # Visualization of the pareto front below ############
    def plot_pf(self, *args, elev=None, azim=None):
        """
        Plots the pareto front.
        """
        if self.n_tasks == 2:
            import matplotlib.pyplot as plt

            fig = plt.figure()
            Y_pareto = np.atleast_2d(
                np.loadtxt("postprocessing/pareto/pf_data_models/y_pareto_output.dat")
            )
            Y_ref = np.atleast_2d(
                np.loadtxt("postprocessing/pareto/pf_data_models/y_reference.dat")
            )

            Y_feasible = np.atleast_2d(
                np.loadtxt("postprocessing/pareto/pf_data_models/y_feasible_output.dat")
            )

            plt.scatter(
                Y_feasible[:, 0], Y_feasible[:, 1], c="gray", label="Feasible points"
            )  # Pareto X1, X2
            plt.scatter(
                Y_pareto[:, 0],
                Y_pareto[:, 1],
                c=Y_pareto[:, 1],
                cmap="rainbow",
                label="Pareto Front",
            )

            if self.pf_optimal_solution:
                Y_ref = np.atleast_2d(
                    np.loadtxt("postprocessing/pareto/pf_data_models/y_reference.dat")
                )
                plt.scatter(
                    Y_ref[:, 0],
                    Y_ref[:, 1],
                    c="pink",
                    marker="X",
                    edgecolor="k",
                    s=400,
                    label="Optimal Pareto Point",
                )

            plt.title("2D Pareto Front")
            if len(args) == 0 or len(args) == 1:
                print("Not enough labels, using default names, task1 and task2")
                label1 = "task1"
                label2 = "task2"
            elif len(args) > 2:
                print("Number of labels greater than the number of tasks")
                label1 = "task1"
                label2 = "task2"
            else:
                label1 = "%s" % (args[0])
                label2 = "%s" % (args[1])
            plt.xlabel("%s" % (label1), fontsize=16)
            plt.ylabel("%s" % (label2), fontsize=16)
            plt.grid()
            plt.xticks(fontsize=16)
            plt.yticks(fontsize=16)
            plt.legend()
            plt.tight_layout()
            plt.savefig("postprocessing/pareto/pf_graph_models/pareto_front_y.png")
            # plt.show()

        if self.n_tasks == 3:
            import matplotlib.pyplot as plt

            data_feas = np.atleast_2d(
                np.loadtxt("postprocessing/pareto/pf_data_models/y_feasible_output.dat")
            )
            data_pf = np.atleast_2d(
                np.loadtxt("postprocessing/pareto/pf_data_models/y_pareto_output.dat")
            )
            x = data_feas[:, 0]
            y = data_feas[:, 1]
            z = data_feas[:, 2]
            x_pf = data_pf[:, 0]
            y_pf = data_pf[:, 1]
            z_pf = data_pf[:, 2]
            fig = plt.figure(figsize=(8, 6))
            ax = fig.add_subplot(111, projection="3d")
            # Customize the scatter plot with colors, sizes, and transparency
            ax.scatter(x, y, z, s=50, color="grey", alpha=0.7)
            ax.scatter(
                x_pf,
                y_pf,
                z_pf,
                s=50,
                c=z_pf,
                cmap="viridis",
                alpha=0.9,
                label="Pareto front",
                zorder=10,
            )

            if self.pf_optimal_solution:
                Y_ref = np.atleast_2d(
                    np.loadtxt("postprocessing/pareto/pf_data_models/y_reference.dat")
                )
                ax.scatter(
                    Y_ref[:, 0],
                    Y_ref[:, 1],
                    Y_ref[:, 2],
                    c="pink",
                    marker="X",
                    edgecolor="k",
                    s=400,
                    label="Optimal Pareto Point",
                    zorder=11,
                )

            # Customize the axis labels and title
            if len(args) in [0, 1, 2]:
                print("Not enough labels, using default names, task1 and task2")
                label1 = "task1"
                label2 = "task2"
                label3 = "task3"
            elif len(args) > 3:
                print("Number of labels greater than the number of tasks")
                label1 = "task1"
                label2 = "task2"
                label3 = "task3"
            else:
                label1 = "%s" % (args[0])
                label2 = "%s" % (args[1])
                label3 = "%s" % (args[2])

            ax.set_xlabel("%s" % label1, fontsize=16, labelpad=10)
            ax.set_ylabel("%s" % label2, fontsize=16, labelpad=10)
            ax.set_zlabel("%s" % label3, fontsize=16, labelpad=10)
            ax.set_title("Pareto front in 3D", fontsize=16)

            # Change tick label font size
            ax.tick_params(axis="x", labelsize=16)
            ax.tick_params(axis="y", labelsize=16)
            ax.tick_params(axis="z", labelsize=16)

            # Show the plot
            # Customize view angle
            ax.view_init(elev=elev, azim=azim)  # Set the elevation and azimuth angles
            # plt.show()
            plt.savefig("postprocessing/pareto/pf_graph_models/3d_pareto_front.png")

    def plot_pos(self, *args, **kwargs):
        """
        Plots the pareto optimal soultions.
        """
        if self.n_dim == 1:
            model_names = []
            for names in kwargs.items():
                model_names.append(names[1])

            import matplotlib.pyplot as plt

            for i in range(self.n_tasks):
                plt.clf()
                data = np.loadtxt("models/grid_model_obj_%s.dat" % i)
                X_p = np.atleast_2d(
                    np.loadtxt(
                        "postprocessing/pareto/pf_data_models/x_pareto_output.dat"
                    )
                )

                if len(args) == 0:
                    print("Not enough label, using default name 'X'")
                    x_label = "X"
                elif len(args) > 1:
                    print("Too many labels")
                    x_label = "X"
                else:
                    x_label = "%s" % args[0]

                if model_names == []:
                    model_label = "Model %s" % i
                else:
                    model_label = "%s" % model_names[i]

                plt.plot(data[:, 0], data[:, 1], lw=4, label=model_label)

                # print(X_p, np.shape((X_p))[1] )
                plt.scatter(X_p, 0 * X_p + np.min(data[:, 1]))

                if self.pf_optimal_solution:
                    X_ref = np.atleast_2d(
                        np.loadtxt(
                            "postprocessing/pareto/pf_data_models/x_reference.dat"
                        )
                    )
                    plt.scatter(
                        X_ref,
                        np.min(data[:, 1]),
                        c="pink",
                        marker="X",
                        edgecolor="k",
                        s=400,
                        label="Optimal Pareto Point",
                        zorder=11,
                    )

                plt.xticks(fontsize=16)
                plt.yticks(fontsize=16)
                plt.xlabel("%s" % x_label, fontsize=16)
                plt.ylabel("%s" % model_label, fontsize=16)
                plt.legend(fontsize=16)
                plt.tight_layout()
                plt.savefig(
                    "postprocessing/pareto/pf_graph_models/pareto_optimal_solution_task_%s.png"
                    % i
                )

        if self.n_dim == 2:
            model_names = []
            for names in kwargs.items():
                model_names.append(names[1])

            import matplotlib.pyplot as plt

            fig = plt.figure()
            X_pareto = np.atleast_2d(
                np.loadtxt("postprocessing/pareto/pf_data_models/x_pareto_output.dat")
            )

            for i in range(self.n_tasks):
                plt.clf()
                data = np.loadtxt("models/grid_model_obj_%s.dat" % i)
                mesh_ = int(np.sqrt(len(data)))
                xp1 = np.linspace(self.bounds[0][0], self.bounds[0][1], mesh_)
                xp2 = np.linspace(self.bounds[1][0], self.bounds[1][1], mesh_)
                y = data[:, 2]
                plt.scatter(
                    X_pareto[:, 0],
                    X_pareto[:, 1],
                    c=X_pareto[:, 1],
                    cmap="rainbow",
                    zorder=5,
                    label="pareto optimal solutions",
                )

                cset = plt.contourf(
                    xp1, xp2, y.reshape(len(xp1), len(xp2)), levels=mesh_, cmap="Greys"
                )
                plt.contour(
                    xp1,
                    xp2,
                    y.reshape(len(xp1), len(xp2)),
                    20,
                    colors="teal",
                    linestyles="dotted",
                    levels=mesh_,
                )

                if self.pf_optimal_solution:
                    X_ref = np.atleast_2d(
                        np.loadtxt(
                            "postprocessing/pareto/pf_data_models/x_reference.dat"
                        )
                    )
                    plt.scatter(
                        X_ref[:, 0],
                        X_ref[:, 1],
                        c="pink",
                        marker="X",
                        edgecolor="k",
                        s=400,
                        label="Optimal Pareto Point",
                        zorder=11,
                    )

                # Add a color bar
                colorbar = plt.colorbar(cset, pad=0.15)
                # Change the size of colorbar tick labels
                colorbar.ax.tick_params(labelsize=16)  # Adjust the tick label siz
                if model_names == []:
                    colorbar.set_label("Model %s" % i, fontsize=16)
                else:
                    colorbar.set_label("%s" % model_names[i], fontsize=16)

                if len(args) == 0 or len(args) == 1:
                    print("Not enough labels, using default names, X1 and X2")
                    label1 = "X1"
                    label2 = "X2"
                elif len(args) > 2:
                    print("Number of labels greater than the number of tasks")
                    label1 = "X1"
                    label2 = "X2"
                else:
                    label1 = "%s" % (args[0])
                    label2 = "%s" % (args[1])
                plt.xlabel("%s" % (label1), fontsize=16)
                plt.ylabel("%s" % (label2), fontsize=16)
                plt.xticks(fontsize=16)
                plt.xticks(
                    np.linspace(self.bounds[0][0], self.bounds[0][1], 5)
                )  # Adjust the range based on your data
                plt.yticks(fontsize=16)
                plt.yticks(np.linspace(self.bounds[1][0], self.bounds[1][1], 5))
                plt.legend()
                plt.tight_layout()
                plt.savefig(
                    "postprocessing/pareto/pf_graph_models/pareto_optimal_solutions_task_%s.png"
                    % (i)
                )

        if self.n_dim == 3:
            import matplotlib.pyplot as plt

            data_inp = np.atleast_2d(
                np.loadtxt("postprocessing/pareto/pf_data_models/x_pareto_output.dat")
            )
            x = data_inp[:, 0]
            y = data_inp[:, 1]
            z = data_inp[:, 2]
            # Create a 3D scatter plot
            fig = plt.figure(figsize=(8, 6))
            ax = fig.add_subplot(111, projection="3d")
            # Customize the scatter plot with colors, sizes, and transparency
            scatter = ax.scatter(
                x, y, z, s=50, c=z, cmap="viridis", alpha=0.7, label="Model"
            )
            if self.pf_optimal_solution:
                X_ref = np.atleast_2d(
                    np.loadtxt("postprocessing/pareto/pf_data_models/x_reference.dat")
                )
                plt.scatter(
                    X_ref[:, 0],
                    X_ref[:, 1],
                    X_ref[:, 2],
                    c="pink",
                    marker="X",
                    edgecolor="k",
                    s=400,
                    label="Optimal Pareto Point",
                    zorder=11,
                )

            # Customize the axis labels and title
            if len(args) == 0 or len(args) == 1 or len(args) == 2:
                print("Not enough labels, using default names, X_i; i  = 1,2,3")
                label1 = "X1"
                label2 = "X2"
                label3 = "X3"
            elif len(args) > 3:
                print("Number of labels greater than the number of tasks")
                label1 = "X1"
                label2 = "X2"
                label3 = "X3"
            else:
                label1 = "%s" % (args[0])
                label2 = "%s" % (args[1])
                label3 = "%s" % (args[2])
            ax.set_xlabel("%s" % label1, fontsize=16, labelpad=10)
            ax.set_ylabel("%s" % label2, fontsize=16, labelpad=10)
            ax.set_zlabel("%s" % label3, fontsize=16, labelpad=10)
            ax.set_title("3D Pareto optimal solutions", fontsize=16)

            # Change tick label font size
            ax.tick_params(axis="x", labelsize=16)  # pad = 10
            ax.tick_params(axis="y", labelsize=16)
            ax.tick_params(axis="z", labelsize=16)

            plt.savefig(
                "postprocessing/pareto/pf_graph_models/3d_pareto_optimal_solutions.png"
            )

            model_names = []
            for names in kwargs.items():
                model_names.append(names[1])

            for i in range(self.n_tasks):
                data_inp = np.loadtxt("models/grid_model_obj_%s.dat" % i)
                x = data_inp[:, 0]
                y = data_inp[:, 1]
                z = data_inp[:, 2]
                out = data_inp[:, 3]
                # Create a 3D scatter plot
                fig = plt.figure(figsize=(8, 6))
                ax = fig.add_subplot(111, projection="3d")

                # Customize the scatter plot with colors, sizes, and transparency
                scatter = ax.scatter(
                    x, y, z, s=50, c=out, cmap="viridis", alpha=0.7, label="Model"
                )

                # Add a color bar
                colorbar = plt.colorbar(scatter, pad=0.15)
                if model_names == []:
                    colorbar.set_label("Model %s" % i, fontsize=16)
                else:
                    colorbar.set_label("%s" % model_names[i], fontsize=16)

                # Customize the axis labels and title
                if len(args) == 0 or len(args) == 1 or len(args) == 2:
                    print("Not enough labels, using default names, X_i; i  = 1,2,3")
                    label1 = "X1"
                    label2 = "X2"
                    label3 = "X3"
                elif len(args) > 3:
                    print("Number of labels greater than the number of tasks")
                    label1 = "X1"
                    label2 = "X2"
                    label3 = "X3"
                else:
                    label1 = "%s" % (args[0])
                    label2 = "%s" % (args[1])
                    label3 = "%s" % (args[2])
                ax.set_xlabel("%s" % label1, fontsize=16, labelpad=10)
                ax.set_ylabel("%s" % label2, fontsize=16, labelpad=10)
                ax.set_zlabel("%s" % label3, fontsize=16, labelpad=10)
                ax.set_title("3D Surrogate model", fontsize=16)

                # Change tick label font size
                ax.tick_params(axis="x", labelsize=16)  # pad = 10
                ax.tick_params(axis="y", labelsize=16)
                ax.tick_params(axis="z", labelsize=16)
                # Change limits on color bar
                scatter.set_clim(np.min(out), np.max(out))

                # Change the size of colorbar tick labels
                colorbar.ax.tick_params(labelsize=16)  # Adjust the tick label siz

                # Show the plot
                # ax.view_init(elev=elev, azim=azim)
                # plt.show()
                plt.savefig(
                    "postprocessing/pareto/pf_graph_models/3d_surrogate_model_%s.png"
                    % i
                )



from __future__ import annotations

import GPy
import numpy as np
from GPy.kern import Kern
from numpy.typing import ArrayLike, NDArray

from boss.bo.models.single_task import SingleTaskModel
from boss.utils.arrays import shape_consistent, shape_consistent_XY


class MultiTaskModel(SingleTaskModel):
    """
    Functionality for creating, refitting and optimizing a multi-task GP model.
    """

    def __init__(
        self,
        kernel: Kern,
        X: ArrayLike,
        Y: ArrayLike,
        noise: float = 1e-12,
        ynorm: bool = False,
        noise_optim=False,
    ) -> None:
        """
        Initializes the multi-task model class.
        """
        self._dim = kernel.input_dim
        self.num_tasks = kernel.parameters[-1].output_dim

        X, Y = shape_consistent_XY(X, Y, x_dim=self.dim)
        inds = np.squeeze(X[:, -1]).astype(int)
        self.check_task_indices(inds)
        XX = [X[inds == index, :-1] for index in range(self.num_tasks)]
        YY = [Y[inds == index] for index in range(self.num_tasks)]

        # normalise observation mean and scale:
        self.normmean = [np.mean(Y) for Y in YY]
        self.normsd = [1] * self.num_tasks

        # scale normalisation is not used unless ynorm is true:
        self.use_norm = ynorm
        self.normsd = [np.ptp(Y) for Y in YY] if self.use_norm else 1

        # normalised observation list:
        YY_norm = [(Y - m) / s for Y, m, s in zip(YY, self.normmean, self.normsd)]
        self.model = GPy.models.GPCoregionalizedRegression(XX, YY_norm, kernel=kernel)
        self.noise_optim = noise_optim
        if not noise_optim:
            self.model.mixed_noise.constrain_fixed(noise)

    @property
    def dim(self) -> int:
        return self._dim

    def get_X(self, index: int | None = None) -> NDArray:
        """
        Returns observed X.
        """
        if index is None:
            return self.model.X
        else:
            return self.model.X[self.inds == index, :-1]

    def get_Y(self, index: int | None = None) -> NDArray:
        """
        Returns observed Y.
        """
        if index is None:
            Y = self.model.Y.copy()
            for index in range(self.num_tasks):
                Y[self.inds == index] *= self.normsd[index]
                Y[self.inds == index] += self.normmean[index]
            return Y
        else:
            Y_norm = self.model.Y[self.inds == index]
            return Y_norm * self.normsd[index] + self.normmean[index]

    @property
    def X(self) -> NDArray:
        return self.get_X()

    @property
    def Y(self) -> NDArray:
        return self.get_Y()

    @property
    def inds(self) -> NDArray:
        return self.model.X[:, -1].astype(int)

    def add_data(self, X_new: ArrayLike, Y_new: ArrayLike) -> None:
        """
        Updates the model evidence (observations) dataset appending.
        """
        X_new, Y_new = shape_consistent_XY(X_new, Y_new, x_dim=self.dim)
        inds_new = X_new[:, -1].astype(int)

        # construct new datasets
        X = np.vstack([self.X, X_new])
        Y = np.vstack([self.Y, Y_new])
        inds = X[:, -1].astype(int)

        # update normalisation
        Y_norm = np.vstack([self.model.Y, np.zeros_like(Y_new)])
        for i in np.unique(inds_new):
            self.normmean[i] = np.mean(Y[inds == i])
            if self.use_norm:
                self.normsd[i] = np.ptp(Y[inds == i])
            Y_norm[inds == i] = (Y[inds == i] - self.normmean[i]) / self.normsd[i]

        # update model
        self.model.Y_metadata = {"output_index": inds}
        self.model.set_XY(X, Y_norm)

    def redefine_data(self, X: ArrayLike, Y: ArrayLike) -> None:
        """
        Updates the model evidence (observations) dataset overwriting.
        """
        X, Y = shape_consistent_XY(X, Y, x_dim=self.dim)
        inds = X[:, -1].astype(int)
        self.check_task_indices(inds)

        # update normalisation
        Y_norm = np.zeros_like(Y)
        for i in range(self.num_tasks):
            self.normmean[i] = np.mean(Y[inds == i])
            if self.use_norm:
                self.normsd[i] = np.ptp(Y[inds == i])
            Y_norm[inds == i] = (Y[inds == i] - self.normmean[i]) / self.normsd[i]

        # update model
        self.model.Y_metadata = {"output_index": inds}
        self.model.set_XY(X, Y_norm)

    def get_best_xy(self, index: int | None = None) -> tuple[NDArray, float]:
        """
        Returns the lowest energy acquisitions (x, y).
        """
        if index is None:
            x_best = []
            y_best = []
            for index in range(self.num_tasks):
                Y_i = self.get_Y(index)
                x_best.append(np.append(self.get_X(index)[np.argmin(Y_i)], index))
                y_best.append(np.min(Y_i))
        else:
            Y_i = self.get_Y(index)
            x_best = np.array(self.get_X(index)[np.argmin(Y_i)])
            y_best = np.min(Y_i)
        return x_best, y_best

    def check_task_indices(self, inds: NDArray) -> None:
        """
        Raises an error if all tasks are not included in the index list or if
        the list includes more tasks than expected.
        """
        counts = np.bincount(inds, minlength=self.num_tasks)
        if not np.all(counts > 0):
            raise ValueError("All tasks must be represented in the dataset.")

        num_tasks = max(inds) + 1
        if num_tasks > self.num_tasks:
            raise ValueError(
                f"Received a dataset with {num_tasks} tasks. "
                f"Expected {self.num_tasks} tasks."
            )

    def extend_input(self, x: ArrayLike, index: ArrayLike) -> NDArray:
        """
        Returns x extended with task index.
        """
        x = np.atleast_2d(x)
        inds = np.full((len(x), 1), np.array(index).reshape(-1, 1))
        x = np.hstack((x, inds))
        return x

    def predict(
        self,
        X: ArrayLike,
        index: ArrayLike | None = None,
        noise: bool = True,
        norm: bool = False,
    ):
        """
        Returns model prediction mean and variance at point x, with or without
        model variance (noise) and normalisation (norm).

        Task index can be included in the input X or provided with index.
        """
        # extend x with task index if needed
        X = shape_consistent(X, self.dim - (index is not None))
        if index is not None:
            X = self.extend_input(X, index)
        # build metadata
        inds = X[:, -1].astype(int)
        meta = {"output_index": inds}
        # predict output
        m, v = self.model.predict(X, Y_metadata=meta, include_likelihood=noise)
        v = np.clip(v, 1e-12, np.inf)
        if norm:
            return m, v
        # remove normalisation
        for i in np.unique(inds):
            m[inds == i] = m[inds == i] * self.normsd[i] + self.normmean[i]
            v[inds == i] = v[inds == i] * self.normsd[i] ** 2
        return m, v

    def predict_grads(
        self, X: ArrayLike, index: ArrayLike | None = None, norm: bool = False
    ) -> tuple[NDArray, NDArray]:
        """
        Returns model prediction mean and variance gradients with respect to
        input at point x, with or without normalisation (norm).

        Task index can be included in the input x or provided with index.
        """
        # extend x with task index if needed
        X = shape_consistent(X, self.dim - (index is not None))
        if index is not None:
            X = self.extend_input(X, index)
        # predictive gradients
        dmdx, dvdx = self.model.predictive_gradients(np.atleast_2d(X))
        if norm:
            return dmdx, dvdx
        # remove normalisation
        inds = X[:, -1].astype(int)
        for i in np.unique(inds):
            dmdx[inds == i] *= self.normsd[i]
            dvdx[inds == i] *= self.normsd[i] ** 2
        return dmdx, dvdx

    def predict_mean_sd_grads(
        self,
        X: ArrayLike,
        index: ArrayLike | None = None,
        noise: bool = True,
        norm: bool = True,
    ) -> tuple[NDArray, NDArray, NDArray, NDArray]:
        """
        Returns the model prediction mean, standard deviation and their
        gradients at point x, with or without model variance (noise) and
        normalisation (norm).

        Task index can be included in the input x or provided with index.
        """
        m, v = self.predict(X, index=index, noise=noise, norm=norm)
        dmdx, dvdx = self.predict_grads(X, index=index, norm=norm)
        dmdx = dmdx[:, :, 0]
        dsdx = dvdx / (2 * np.sqrt(v))
        return m, np.sqrt(v), dmdx, dsdx

    def predict_mean_grad(
        self, X: ArrayLike, index: ArrayLike | None = None, norm: bool = True
    ) -> tuple[NDArray, NDArray]:
        """
        Returns model mean and its gradient at point x, with or without
        normalisation (norm).

        Task index can be included in the input x or provided with index.
        """
        m, _ = self.predict(X, index=index, norm=norm)
        dmdx, _ = self.predict_grads(X, index=index, norm=norm)
        return m, dmdx

    def estimate_num_local_minima(self, search_bounds: ArrayLike) -> int:
        """
        Returns estimated number of local minima calculated based on model
        properties.
        """
        # For the ith dimension, the number of local minima along a slice
        # is approximately n(i) = boundlength(i)/(2*lengthscale(i)). Note
        # that periodic kernels operate on normalised distances: distance
        # between inputs that are period(i)/2 apart is 1. To get the total
        # number of minima for all of the search space, multiply together
        # n(i) over all i.
        numpts = 1

        # get baseline kernel parameters (exclude coregionalisation kernel)
        ks = self.model.kern.parameters[:-1]
        for bounds, kern in zip(search_bounds, ks):
            if hasattr(kern, "period"):
                bound_distance = (bounds[1] - bounds[0]) / kern.period[0]
            else:
                bound_distance = (bounds[1] - bounds[0]) / 2
            numpts *= max(1, bound_distance / kern.lengthscale[0])
        return int(numpts)

    def predict_task_covariance(self, x: ArrayLike) -> NDArray:
        """
        Return predictive covariance between tasks at point x.
        """
        inds = np.arange(self.num_tasks)
        x = np.squeeze(x)[:-1]
        x_list = np.vstack([self.extend_input(x, i) for i in inds])
        meta = {"output_index": inds.astype(int)}
        _, cov = self.model.predict(x_list, Y_metadata=meta, full_cov=True)
        return np.outer(self.normsd, self.normsd) * cov

    def get_task_covariance(self) -> NDArray:
        """
        Returns estimated task covariance matrix.
        """
        kappa = np.array(self.model.kern.parameters[-1].kappa)
        W = np.array(self.model.kern.parameters[-1].W)
        cov = np.outer(W, W) + np.diag(kappa)
        return np.outer(self.normsd, self.normsd) * cov



from __future__ import annotations
from mimetypes import init

from typing import Any, Callable

import numpy as np
from numpy.typing import ArrayLike, NDArray

import boss.bo.factory as factory
import boss.io.ioutils as ioutils
import boss.io.parse as parse
from boss.bo.initmanager import InitManager
from boss.bo.models.model import Model
from boss.bo.results import BOResults, minimize_model
from boss.bo.rstmanager import RstManager
from boss.io.main_output import MainOutput
from boss.settings import Settings
from boss.utils.arrays import shape_consistent_XY
from boss.utils.minimization import Minimization
from boss.utils.timer import Timer


class BOMain:
    """
    Class for handling Bayesian Optimization
    """

    def __init__(self, f: Callable, bounds: NDArray, **keywords: Any) -> None:
        keywords["bounds"] = bounds
        settings = Settings(keywords, f=f)
        self.settings = settings
        self.rst_manager = RstManager(settings)
        self._setup()

    @classmethod
    def from_file(
        cls,
        ipfile: str,
        outfile: str | None = None,
        f: Callable | None = None,
        **new_keywords: Any,
    ) -> "BOMain":
        """Initialize BOMain from a BOSS input or rst file.

        Parameters
        ----------
        ipfile : path_like
            The input file to initialize from, can be either
            a boss input or rst file.
        **new_keywords
            Any new BOSS keywords.
        """
        self = cls.__new__(cls)
        input_data = parse.parse_input_file(ipfile)
        rst_data = input_data.get("rst_data", np.array([]))
        keywords = input_data.get("keywords", {})
        keywords.update(new_keywords)
        self.settings = Settings(keywords, f=f)
        self.rst_manager = RstManager(self.settings, rst_data)
        cls._setup(self)
        return self

    @classmethod
    def from_settings(
        cls, settings: Settings, rst_data: NDArray | None = None
    ) -> "BOMain":
        """Construction from a Settings object."""
        self = cls.__new__(cls)
        self.settings = settings
        self.rst_manager = RstManager(self.settings, rst_data)
        cls._setup(self)
        return self

    def _setup(self) -> None:
        """Common setup for all factory methods."""
        settings = self.settings
        self.main_output = MainOutput(settings)

        # Resolve optimisation domain
        if settings.is_multi:
            self.dim = settings.dim + 1
            self.bounds = np.vstack((settings["bounds"], [[0, 0]]))
        else:
            self.dim = settings.dim
            self.bounds = settings["bounds"]

        self.y_dim = 1 + self.dim * self.settings.is_grad

        self.user_func = settings.f
        self._model = None

        self.acqfn = factory.get_acq_func(settings)
        self.acq_manager = factory.get_acq_manager(settings, self.acqfn)

        Minimization.set_parallel(self.settings["parallel_optims"])

        self.results = BOResults(settings=settings)
        self.results.add_defaults()
        self.conv_checker = factory.get_conv_checker(settings)
        self.itr_curr = 0
        self.cum_cost = 0

    @property
    def model(self) -> Model:
        if self._model is not None:
            return self._model
        else:
            raise AttributeError('Model is not set')

    @model.setter
    def model(self, new_model: Model) -> None:
        self._model = new_model

    def init_model(
        self, X: ArrayLike, Y: ArrayLike, params: ArrayLike | None = None
    ) -> None:
        """Initializes the GP model."""
        X, Y = shape_consistent_XY(X, Y, x_dim=self.dim, y_dim=self.y_dim, nan_pad=False)
        if self.settings["yrange"] is None:
            self.settings["yrange"] = [np.min(Y), np.max(Y)]

        self.model = factory.get_model(self.settings, X, Y)

        # Look for optimized model params from:
        # func arguments -> restart manager -> optimize model
        if params is None:
            params = self.rst_manager.get_params(X.shape[0] + 1)
        else:
            params = np.atleast_1d(params)

        if params is not None:
            self.model.set_unfixed_params(params)
        else:
            if self.settings["initupdate"]:
                self.model.optimize(self.settings["updaterestarts"])

        # connect acqfn to model
        self.acqfn.model = self.model

    def init_run(
        self, X_init: ArrayLike, Y_init: ArrayLike | None = None
    ) -> NDArray:
        """Evalutes initial points and calls model initialization.

        This method assumes that, at least, initial X-values have been determined,
        i.e. passed by the user or retrieved from file or an InitManager
        (see resolve_initpts).

        Parameters
        ----------
        X_init: np.ndarray
            Initial X-values, must be completely specified,
            i.e., cannot be None nor have nan-elements.

        Y_init: Optional[np.ndarray]
            Initial Y-values, can be omitted entirely or partially, in which
            case the user function will be evaluated at the missing X-locations.

        Returns
        -------
        X_next: np.ndarray
            The very first acquisition that will be used in the BO-run.
        """
        self.main_output.new_file()
        self.rst_manager.new_file()
        self.results.clear()

        # Ensure X,Y initial values are 2D arrays of the same size.
        x_dim = self.dim
        X_init, Y_init = shape_consistent_XY(
            X_init, Y_init, x_dim=x_dim, y_dim=self.y_dim, nan_pad=True
        )
        initpts = X_init.shape[0]
        self.settings["initpts"] = initpts
        X = np.empty((0, x_dim), float)
        Y = np.empty((0, self.y_dim), float)
        for i in range(self.settings["initpts"]):
            with self.main_output.summarize_results(self.results):
                # Evaluate the userfn if initial y-data is not available
                if np.isnan(Y_init[i]).any():
                    XY_out = self._eval_user_func(X_init[i, :], write_rst=False) #the user may define to return multiple points per input point
                    X_new = XY_out[0]
                    Y_new = XY_out[1]
                    X = np.vstack((X, X_new))
                    Y = np.vstack((Y, Y_new))
                    for xi, yi in zip(X_new, Y_new): #in order to account for repeated/multiple entries per input point
                        self.rst_manager.write_data(xi, yi)

                else:
                    X = np.vstack((X, X_init[i, :]))
                    Y = np.vstack((Y, Y_init[i, :]))
                    self.rst_manager.write_data(X[i, :], Y[i, :])
                
                # For the first n - 1 initpts the only results are (x,y)-data.
                if i < initpts - 1:
                    self.results.update({"X": X, "Y": Y})

                # For the final initpt: initialize model and calc. all results.
                else:
                    self.init_model(X, Y)
                    labeled_params = self.model.get_all_params(include_fixed=False)
                    self.rst_manager.write_column_labels(labeled_params)
                    self.rst_manager.write_model_params(self.model.get_unfixed_params())

                    # Get the next acquisition and update all results.
                    X_next = self.acquire()
                    self._update_results(X_next)

        self.results.set_num_init_batches(initpts)
        return X_next

    def get_initpts(self) -> tuple[NDArray, NDArray]:
        """If initial data is not provided, get it from the rst/init manager."""
        if self.rst_manager.data.shape[0] > 0:
            X_init = self.rst_manager.X
            Y_init = self.rst_manager.Y
        else:
            if self.settings.is_multi:
                initpts = self.settings["task_initpts"]
            else:
                initpts = self.settings["initpts"]

            init_manager = InitManager(
                inittype=self.settings["inittype"],
                bounds=self.settings["bounds"],
                initpts=initpts,
                seed=self.settings["seed"],
                scramble=self.settings["initscramble"],
            )
            X_init = init_manager.get_all()
            Y_init = (
                np.empty((len(X_init), 1 + self.settings.is_grad * self.dim)) * np.nan
            )
        return X_init, Y_init

    def run(
        self,
        X_init: ArrayLike | None = None,
        Y_init: ArrayLike | None = None,
        iterpts: int | None = None,
        maxcost: float | None = None,
    ) -> BOResults:
        """
        The Bayesian optimization main loop. Evaluates first the initialization
        points, then creates a GP model and uses it and an acquisition function
        to locate the next points where to evaluate. Stops when a pre-specified
        number of initialization points and BO points have been acquired or a
        convergence criterion or cost limit is met.

        Parameters
        ----------
        X_init : Optional[ArrayLike]
            Initial input points, provided as rows in a 2D array, to use for the BO.
            Will take precedence over any other specification of initial points coming
            from initmanagers or restart files.
        Y_init : Optional[ArrayLike]
            Precomputed output data corresponding to X_init. Can contain fewer data
            points than X_init, in which case the userfn will be called instead.
        iterpts : Optional[int]
            The maximum number of BO iterations that will be performed unless another
            termination criterion is fulfilled first. This option, if set, will
            take precedence over the iterpts specified in settings.
        maxcost : Optional[float]
            A cost limit for user function evaluations that, if provided,
            will be used to terminate the BO-loop.

        Returns
        -------
        BOResults
            Provides access to the most important results from the optimization.
        """
        if iterpts:
            self.settings["iterpts"] = iterpts
        if maxcost:
            self.settings["maxcost"] = maxcost
        # Provide the model with data and create output files if we're
        # starting fresh or restarting from file.
        if self._model is None:
            # Handle initial points
            if X_init is not None:
                X_init, Y_init = shape_consistent_XY(
                    X_init, Y_init, x_dim=self.dim, nan_pad=True, y_dim=self.y_dim
                )
            else:
                X_init, Y_init = self.get_initpts()

            X_next = self.init_run(X_init, Y_init)
        # The model already has data and we just need the next acquisition.
        else:
            X_next = self.acquire()

        # BO main loop
        for _ in range(self.itr_curr, self.settings["iterpts"] + 1):
            if self.exceeds_cost_limit(X_next):
                self.main_output.maxcost_stop()
                break

            with self.main_output.summarize_results(self.results):
                # 1. User func evaluation
                X_new, Y_new = self._eval_user_func(X_next)

                # 2. Model update: refit model to the updated data & optimize hyperparams.
                self._update_model(X_new, Y_new)

                # 3. Get a new acquisition for the next iteration.
                X_next = self.acquire()

                # 4. Update results and write a summary to the outfile.
                self._update_results(X_next)

            # 5. Convergence check
            if self.conv_checker is not None and self.conv_checker(self.results):
                self.main_output.convergence_stop()
                break

        return self.results

    def acquire(self) -> NDArray:
        """Gets the next acquisition."""
        X_next = self.acq_manager.acquire()
        if self.acq_manager.message:
            self.main_output.progress_msg(self.acq_manager.message)
        return X_next

    def _eval_user_func(
        self, X: NDArray, write_rst: bool = True, write_out: bool = True
    ) -> tuple[NDArray, NDArray]:
        """Evalutes the userfn and writes data to the rstfile."""
        timer = Timer()
        with timer.time():
            func_out = self.user_func(X)
        X_out = func_out.X
        Y = func_out.Y

        # Write info to file
        for x_out, y in zip(X_out, Y):
            if write_rst:
                self.rst_manager.write_data(x_out, y)
            if write_out:
                self.main_output.progress_msg(
                    "Evaluating objective function at x ="
                    + ioutils.oneDarray_line(x_out, len(x_out), float),
                )
        self.main_output.progress_msg(
            f"Objective function evaluated, time [s] {timer.lap_time}",
        )

        # Update accumulated cost if needed
        if self.settings["maxcost"] is not None:
            self.cum_cost += self.acqfn.evaluate_cost(X_out)

        return X_out, Y

    def _update_model(self, X_next: NDArray, Y_next: NDArray) -> None:
        """Optimizes and refits model with new data."""

        # Store new data
        self.model.add_data(X_next, Y_next)

        # Optimize model if needed.
        updatefreq = self.settings["updatefreq"]
        updaterestarts = self.settings["updaterestarts"]
        updateoffset = self.settings["updateoffset"]

        itr1 = self.itr_curr - 1  # start optimizing on iteration 1
        should_optimize = (
            updatefreq > 0 and itr1 >= updateoffset and (itr1 % updatefreq == 0)
        )
        if should_optimize:
            self.model.optimize(updaterestarts)

        # Add model parameters to rst-file.
        self.rst_manager.write_model_params(self.model.get_unfixed_params())

    def _update_results(self, X_next: NDArray) -> None:
        """Updates BOResults with acqs, model and min info."""
        minfreq = self.settings["minfreq"]
        self.results.update(
            {
                "X": self.model.X,
                "Y": self.model.Y,
                "X_next": X_next,
                "model_params": self.model.get_unfixed_params(),
            }
        )
        # Add global min. info
        should_optimize = (
            (minfreq > 0 and (self.itr_curr % minfreq == 0))
            or self.itr_curr == self.settings["iterpts"]  # max iters reached
            or self.exceeds_cost_limit(X_next)  # max cost reached
        )

        if should_optimize:
            x_glmin, mu_glmin, nu_glmin = minimize_model(
                self.model,
                self.bounds,
                self.settings["optimtype"],
                self.settings["kernel"],
                self.settings["min_dist_acqs"],
                accuracy=self.settings["minzacc"],
            )
            self.results.update(
                {
                    "x_glmin": x_glmin,
                    "mu_glmin": mu_glmin,
                    "nu_glmin": nu_glmin,
                }
            )
        self.results.update({"t_total": self.main_output.timer.getTotalTime()})
        self.itr_curr += 1

    def exceeds_cost_limit(self, X_next: NDArray) -> bool:
        """
        Checks whether the next acquisition would exceed cost limit.
        """
        if self.settings["maxcost"] is None:
            return False
        else:
            cost_next = self.acqfn.evaluate_cost(X_next)
            return self.cum_cost + cost_next > self.settings["maxcost"]



from __future__ import annotations

import datetime
import itertools
import logging
import socket
from pathlib import Path
from typing import Any

import numpy as np
from numpy.typing import NDArray

from boss import __version__
from boss.bo.bo_main import BOMain
from boss.bo.initmanager import InitManager
from boss.testing.db import BenchmarkDatabase, ColumnSpec, get_callable_name
from boss.utils.arrays import concatenated_cartesian_product


class KeywordMix:
    """Enables combining keywords and values using a set of common operations.
    """
    def __init__(self, name: str, values: list | tuple | np.ndarray) -> None:
        """
        Creates an atomic keyword-values pair.

        Parameters
        ----------
        name : str
            Name of the keyword.
        values : list | tuple | np.ndarray
            Keywords values.
        """
        self.names = [name]
        self.values = [values]
        self.indices = np.arange(0, len(values))[:, None]

    def __iter__(self) -> KeywordMixIterator:
        return KeywordMixIterator(self)

    def __len__(self) -> int:
        """Returns the number of encoded keyword combinations."""
        return len(self.indices)

    def chain(self, other: KeywordMix, inplace: bool = False) -> KeywordMix:
        new = self if inplace else self.__new__(type(self))

        name_overlap = set(self.names) & set(other.names)
        shape1 = self.indices.shape
        shape2 = other.indices.shape
        inds = -1 * np.ones(
            (shape1[0] + shape2[0], shape1[1] + shape2[1] - len(name_overlap)),
            dtype=np.int64
        )
        inds[:shape1[0], :shape1[1]] = self.indices
        if not name_overlap:
            inds[shape1[0]:, shape1[1]:] = other.indices
            new.values = self.values + other.values
        else:
            new.values = self.values
            for name in name_overlap:
                i = self.names.index(name)
                j = other.names.index(name)
                inds[shape1[0]:, i] = other.indices[:, j] + len(self.values[i])
                new.values[i] += other.values[j]

            for i, name in enumerate(set(other.names) - name_overlap):
                j = other.names.index(name)
                new.values.append(other.values[j])
                inds[shape1[0]:, shape1[1] + i] = other.indices[:, j]
            
        new.indices = inds
        new.names = list(set(self.names) | set(other.names))
        return new

    def zip(self, other: KeywordMix, inplace: bool = False) -> KeywordMix:
        new = self if inplace else self.__new__(type(self))
        new.indices = np.hstack((self.indices, other.indices))
        new.names = self.names + other.names
        new.values = self.values + other.values
        return new

    def product(self, other: KeywordMix, inplace: bool = False) -> KeywordMix:
        new = self if inplace else self.__new__(type(self))
        new.indices = concatenated_cartesian_product(self.indices, other.indices)
        new.names = self.names + other.names
        new.values = self.values + other.values
        return new

    def __add__(self, other: KeywordMix) -> KeywordMix:
        return self.zip(other)

    def __iadd__(self, other: KeywordMix) -> KeywordMix:
        return self.zip(other, inplace=True)

    def __truediv__(self, other: KeywordMix) -> KeywordMix:
        return self.chain(other)

    def __itruediv__(self, other: KeywordMix) -> KeywordMix:
        return self.chain(other, inplace=True)

    def __mul__(self, other: KeywordMix) -> KeywordMix:
        return self.product(other)

    def __imul__(self, other: KeywordMix) -> KeywordMix:
        return self.product(other, inplace=True)

    def __str__(self) -> str:
        """Returns a table displaying all encoded keyword combinations."""
        cols = self.names
        vals = [
            [str(self.values[i][k]) if k != -1 else None for i, k in enumerate(ind)]
            for ind in self.indices
        ]
        vals = np.array(vals, dtype=str)
        max_len = lambda col: max(len(c) for c in col)
        widths = np.apply_along_axis(max_len, 0, vals)
        widths = np.max([widths, [len(c) for c in cols]], axis=0)
        tot_width = sum(widths) + 3 * len(cols) + 1

        ruler = f"{'-' * tot_width}\n"
        left = "| "
        right = " |\n"
        middle = " | ".join([f"{c: <{widths[i]}}" for i, c in enumerate(cols)])
        header = left + middle + right
        body = ""
        for row in vals:
            middle = " | ".join([f"{x: <{widths[i]}}" for i, x in enumerate(row)])
            middle = middle.replace("None", f"    ")
            body += left + middle + right

        table = ruler + header + ruler + body + ruler
        return table


class KeywordMixIterator:
    def __init__(self, keyword_mix: KeywordMix) -> None:
        self.names = keyword_mix.names
        self.values = keyword_mix.values
        self.indices = iter(keyword_mix.indices)

    def __next__(self) -> dict[str, Any]:
        inds = next(self.indices)
        vals = [self.values[i][k] if k != -1 else None for i, k in enumerate(inds)]
        kws = {n: v for n, v in zip(self.names, vals) if v is not None}
        return kws


class InitCycler:
    """Cycles through a fixed sequence of randomized initial points.

    Used to make benchmarking comparisons fair: if we have 2 different keyword combinations
    that are each repeated N times, we would like to use the same sequence of N
    randomized initial points for each combination.
    """
    def __init__(self, n_repeats: int) -> None:
        self.entropies = itertools.cycle(
            np.random.SeedSequence().entropy for _ in range(n_repeats)
        )

    def next_initpts(self, settings) -> NDArray:
        seed_seq = np.random.SeedSequence(next(self.entropies))
        im = InitManager(
            seed=seed_seq,
            inittype=settings["inittype"],
            bounds=settings["bounds"],
            initpts=settings["initpts"],
            scramble=settings["initscramble"],
        )
        X_init = im.get_all()
        return X_init


def timestamp() -> str:
    """Returns the current date and time in isoformat"""
    return datetime.datetime.now().replace(microsecond=0).isoformat()


class Benchmarker:
    def __init__(
        self,
        name: str = "benchmark",
        fixed_keywords: dict[str, Any] | None = None,
        variable_keywords: KeywordMix | None = None,
        cycle_initpts: bool = True,
        save_files: bool = True,
        db: BenchmarkDatabase | str | Path | None = None,
        custom_columns: dict[str, ColumnSpec] | None = None,
        loglevel: int = logging.DEBUG,
    ) -> None:
        """Tool for running repeated BOSS runs with different keyword combinations.

        Parameters
        ----------
        name : str
            Name of the benchmarking run, will be used for the database, logfile and
            output file directory.
        fixed_keywords : dict | None
            BOSS keywords that will remain fixed for all runs.
        variable_keywords : KeywordMix
            Keywords that will change for each run, see documentation for the
            KeywordMix object.
        cycle_initpts : bool
            Enables a consistent treatment of initial points for benchmarking:
            a fixed sequence of seeds will be created (at the start of the benchmarking)
            for the prng used by BOSS' InitManager. For example, if  (k1, k2) are two sets
            of keywords and we are doing three repetitions for each set, three arrays of
            initial points (X1, X2, X3) will be created and the six total runs are:
            k1: repeat with X1, X2, X3 -> k2: repeat with X1, X2, X3.
        save_files : bool
            Whether to save boss.out and boss.rst files. If truthy the files
            will be stored in a new directory called {name}_out. The naming convention
            for the files is boss-k-r.{rst,out} where k = index of keyword combination
            and r = index of repetition (relative to current keywords).
        db : BenchmarkDatabase | str | None
            Name or instance of a BenchmarkDatabase object that will track
            the keywords and results of the benchmarking. If none is given a new
            database called name.db will be created.
        custom_columns : dict[str, ColumnSpec] | None
            Custom columns to include in the database, specified via a dict mapping
            column names to ColumnSpec instances. For example:
            mycol = ColumnSpec(int, get_value) will create an integer
            column called mycol whose value is set using the get_value function.
            This function must be defined by the user and is called with Settings
            and row_id as argument.
        loglevel : int
            Logging level of the logger. The logging.DEBUG corresponds to the maximum verbosity.
        """
        self.name = name
        self.fixed_keywords = fixed_keywords or {}
        self.variable_keywords = variable_keywords
        self.logger = logging.getLogger(__name__)
        self.logger.setLevel(loglevel)
        self.n_repeats = 1
        self.save_files = save_files
        self.save_dir = f"{name}_out"
        self.cycle_initpts = cycle_initpts
        self.db = None
        custom_columns = custom_columns or {}

        # Resolve db name.
        db_name = Path()
        if db is None:  # use default db name
            db_name = Path(f"{self.name}.db")
        elif isinstance(db, (str, Path)):  # use given db name
            db_name = Path(db)
        else:  # use given db instance
            self.db = db

        # If we weren't passed a db instance create a new one or load existing.
        if self.db is None:
            if not db_name.is_file():  # create new db instance
                # Only store the BOSS keywords appearing in fixed/variable keywords
                # in the database
                keyword_columns = set(self.fixed_keywords.keys())
                if variable_keywords:
                    keyword_columns.update(variable_keywords.names)
                keyword_columns -= {"f"}
                self.db = BenchmarkDatabase(
                    db_name, keyword_columns=keyword_columns, **custom_columns
                )
            else:  # load old db
                self.db = BenchmarkDatabase.from_file(db_name, **custom_columns)

    def _init_log(self) -> None:
        """Writes initial info to the logfile."""
        logger = self.logger
        fh = logging.FileHandler(f"{self.name}.log", mode="w")
        formatter = logging.Formatter("%(message)s")
        fh.setFormatter(formatter)
        logger.addHandler(fh)

        logger.info("=========================")
        logger.info(f"    BOSS BENCHMARKER    ")
        logger.info("=========================")
        logger.info(f"timestamp: {timestamp()}")
        logger.info(f"BOSS version: {__version__}")
        logger.info(f"host: {socket.gethostname()}")
        logger.info(f"directory: {Path.cwd()}")

        logger.info("\nfixed keywords:")
        for key, val in self.fixed_keywords.items():
            if callable(val):
                val = get_callable_name(val)
            logger.info(f"{'':>4}{key}: {val}")

        logger.info("\nvariable keyword spans:")
        if self.variable_keywords is not None:
            for i, name in enumerate(self.variable_keywords.names):
                vals = self.variable_keywords.values[i]
                vals = [get_callable_name(v) if callable(v) else v for v in vals]
                logger.info(f"{'':>4}{name}: " + " ".join([str(v) for v in vals]))

        logger.info(f"repetitions per keyword combination: {self.n_repeats}")

    def run(self, n_repeats: int = 1) -> BenchmarkDatabase:
        """Runs the benchmarker for the current (mix of variable) keywords.

        Parameters
        ----------
        n_repeat : int
            The number of repeated runs to do for each combination of keyword values.
        """
        self.n_repeats = n_repeats
        if self.save_files:
            Path(self.save_dir).mkdir(exist_ok=True)

        logger = self.logger
        self._init_log()
        logger.info("\nCOMMENCING BECHMARK")
        logger.info("=========================")

        if not self.variable_keywords:
            self._run_repeat()
        else:
            for i, kws in enumerate(self.variable_keywords):
                logger.info(
                    f"Variable keyword combination {i + 1}/{len(self.variable_keywords)}:"
                )
                msg = ""
                for key, val in kws.items():
                    if callable(val):
                        val = get_callable_name(val)
                    msg += f"{'':>4}{key}: {val}"
                logger.info(msg)
                self._run_repeat(var_kws=kws)
                logger.info("-------------------------")

        return self.db

    def _run_repeat(self, var_kws: dict[str, Any] | None = None) -> None:
        """Carries out the actual repeated BOSS runs.

        Parameters
        ----------
        var_kws : dict | None
            Variable keywords that are given to BOSS in addition to the fixed keywords.
        """
        logger = self.logger
        logger.info(f"repetitions:")
        var_kws = var_kws or {}
        all_kws = {**self.fixed_keywords, **var_kws}
        kws_rec = self.db.add_keywords(all_kws)
        init_cycler = None
        if self.cycle_initpts:
            init_cycler = InitCycler(self.n_repeats)

        for i_rep, rep in enumerate(range(self.n_repeats)):
            logger.info(f"{'':>4}{rep + 1}: started {timestamp()}")

            if self.save_files:
                all_kws["outfile"] = str(
                    Path(self.save_dir) / f"boss-{self.db._id_last}-{i_rep}.out"
                )
                all_kws["rstfile"] = str(
                    Path(self.save_dir) / f"boss-{self.db._id_last}-{i_rep}.rst"
                )

            bo = BOMain(**all_kws)
            X_init = None
            if init_cycler:
                X_init = init_cycler.next_initpts(bo.settings)
            bo_res = bo.run(X_init=X_init)
            self.db.add_results(bo_res, keywords=kws_rec)



from __future__ import annotations

import os.path

import numpy as np
from numpy.typing import ArrayLike, NDArray

import boss.io.dump as dump
import boss.io.ioutils as io
from boss.settings import Settings


def split_rst_data(
    rst_data, x_dim: int, y_dim: int = 1
) -> tuple[NDArray, NDArray]:
    """Split the restart data into acquisitions and model parameters.

    Parameters
    ----------
    rst_data: np.ndarray
        The rst_data containing rows of (X, y, theta),
        e.g., as returned by parse_input_data.
    dim: int
        The dimension of the X-data.

    Returns
    -------
    acqs: np.ndarray
        Array containing the iteration number (starting from 1) and the correspoding
        acquisitions found in the restart data.
    mod_par: np.ndarray
        Array containing the iteration number (starting from 1) and the correspoding
        model parameters found in the restart data.
    """
    totalpts = rst_data.shape[0]
    # The number of input pts w/o mod param values is the number of rows with at most
    # x_dim + y_dim entries that are not np.nan.
    inputpts = np.sum(np.sum(~np.isnan(rst_data), axis=1) <= x_dim + y_dim)
    acqs = np.c_[np.arange(1, totalpts + 1), rst_data[:, : x_dim + y_dim]]
    if totalpts > inputpts:
        mod_par = np.c_[
            np.arange(inputpts + 1, totalpts + 1),
            rst_data[inputpts:, x_dim + y_dim :],
        ]
    else:
        mod_par = np.array([])
    return acqs, mod_par


class RstManager:
    """
    A class that handles restart-files (rst-files). These files can be used to
        1) introduce acquisition data from another source as initial values,
        2) continue a run that has been interrupted for some reason,
        3) continue a finished run by acquiring more points or
        4) keep the acquisitions but change the model or settings for a rerun.
    """

    def __init__(self, settings: Settings, rst_data: NDArray | None = None) -> None:
        """
        Initializes the class with an array (settings.rstvals) containing the data
        read from an rst-file.
        """
        if rst_data is None:
            rst_data = np.array([])
        self.data = rst_data
        self.settings = settings
        self.x_dim = settings.dim + int(settings.is_multi)
        self.y_dim = 1 + self.x_dim * self.settings.is_grad
        self.is_grad = settings.is_grad
        self.rstfile = settings["rstfile"]

    @property
    def X(self) -> NDArray | None:
        if self.data.shape[0] > 0:
            return self.data[:, : self.x_dim]
        else:
            return None

    @property
    def Y(self) -> NDArray | None:
        if self.data.shape[0] > 0:
            return self.data[:, self.x_dim : self.x_dim + self.y_dim]
        else:
            return None

    def new_file(self) -> None:
        if os.path.isfile(self.rstfile):
            print("warning: overwriting file '" + self.rstfile + "'")
        keywords = {
            k: v for k, v in self.settings.items() if k in self.settings.user_keys
        }
        dump.dump_input_file(self.rstfile, keywords, results_header=True)
        self.write_column_labels()

    def get_x(self, i: int) -> NDArray | None:
        """
        Returns the i:th acquisition location from the rst-data or None if
        it can't be found.
        """
        if self.data.shape[0] > i:
            x = self.data[i, : self.x_dim]
            return x
        else:
            return None

    def get_y(self, i: int) -> None | NDArray | tuple[NDArray, NDArray]:
        """
        Returns the i:th acquisition evaluation from the rst-data or None if it can't be found.
        """
        if self.data.shape[0] <= i:
            return None

        x_dim = self.x_dim
        y_dim = self.y_dim
        y = self.data[i, x_dim : x_dim + y_dim]

        if np.any(np.isnan(y)):
            return None

        if self.settings.is_grad:
            y, dy = y[x_dim], y[x_dim + 1 : x_dim + y_dim]
            return (y, dy)
        else:
            return y

    def get_params(self, i: int) -> NDArray | None:
        """
        Returns the model paramters at iteration i from the rst-data or None
        if they can't be found.
        """
        if self.data.shape[0] <= i:
            return None

        params = self.data[i, self.x_dim + self.y_dim]
        if np.any(np.isnan(params)):
            return None

        return params

    def write_data(self, x, y):
        """
        Outputs a new data point (x,y) to rst file.
        """
        with open(self.rstfile, "a") as rst:
            rst.write("\n" + io.data_line(x, y, fstr="%23.15E")[:-1])

    def write_column_labels(
        self, labeled_params: dict[str, NDArray] | None = None
    ) -> None:
        """
        Adds a comment with column labels for x, y data and hyperparameters.

        Updates the RESULTS section with a comment containing aligned labels
        x1, x2, .., y1, y2, ..., param1, param2, ... for each column.

        Parameters
        ----------
        labeled_params : dict[str, NDArray] | None
            A dictionary mapping parameter names to values. We typically get
            this from model.get_all_params().
        """
        x_labels = [f"x{i + 1}" for i in range(self.x_dim)]
        y_labels = [f"y{i + 1}" for i in range(self.y_dim)]
        big_width = 23
        small_width = 6
        small_sep = f"{'':{small_width}}"
        line = "#"
        line += "".join([f"{lab:^{big_width}}" for lab in x_labels])[1:] + small_sep
        line += "".join([f"{lab:^{big_width}}" for lab in y_labels])

        if labeled_params is not None:
            line += small_sep
            for label, params in labeled_params.items():
                n_par = params.size
                if n_par > 1: #  add param index to label
                    for i_par in range(n_par):
                        label_i = f"{label}{i_par + 1}"
                        line += f"{label_i:<{big_width}}"
                else:
                    line += f"{label:<{big_width}}"

        with open(self.rstfile, "a") as rst:
            rst.write("\n" + line)

    def write_model_params(self, mod_param):
        """
        Outputs a new set of model parameters to rst file.
        """
        with open(self.rstfile, "a") as rst:
            rst.write("     " + io.data_line(mod_param, fstr="%23.15E")[:-1])



import contextlib
import os
import shutil
from pathlib import Path

import numpy as np


def remove_boss_files():
    path = Path.cwd()
    for file_path in [path / "boss.in", path / "boss.out", path / "boss.rst"]:
        if file_path.exists():
            file_path.unlink()
    shutil.rmtree(path / "postprocessing", ignore_errors=True)


class execute_in(contextlib.ContextDecorator):
    def __init__(self, rundir):
        self.rundir = Path(rundir)
        self.origin = Path.cwd()
        super().__init__()

    def __enter__(self):
        os.chdir(self.rundir)

    def __exit__(self, *exc):
        os.chdir(self.origin)



"""Data output interface for Bayesian optimization results.

This module provides the `Dumper` class, which serves as the data export interface
for post-processing BOSS optimization results. It is typically used by the `PPMain`
object (defined in `boss/pp/pp_main.py`) to write numerical data to text files for
external analysis and archival purposes.

The module is part of the post-processing infrastructure alongside:
- `boss.pp.pp_main.PPMain`: Main post-processing orchestrator that coordinates
  data dumping and plotting
- `boss.pp.plotter.Plotter`: Creates visualization plots from the same results
- `boss.pp.mesh.Mesh`: Manages coordinate grids for function evaluation

The `Dumper` class exports various types of optimization data to human-readable
text files with descriptive headers:
- Hyperparameter evolution over iterations
- Gaussian process model predictions (mean and variance) on grids
- Acquisition function values on grids
- True objective function values (when available)
- Convergence diagnostics
- Predicted minimum locations and values

Typical Usage
-------------
The `Dumper` is usually created and used via `PPMain`:

>>> from boss.pp.pp_main import PPMain
>>> pp = PPMain(results, pp_models=True, pp_acq_funcs=True)
>>> pp.run()

This creates the `postprocessing/` directory with data files alongside plots.

Direct usage is also supported:

>>> from boss.io.dumper import Dumper
>>> dumper = Dumper.from_grid_slice(results)
>>> dumper.dump_model("model_output.dat", itr=-1)
>>> dumper.dump_hyperparams("hyperparams.dat")

Output Format
-------------
All output files are text files with:
- Header lines starting with '#' describing the data format
- Column-oriented data with scientific notation (default precision: 9 decimal places)
- Special formatting for iteration numbers (integer) and ensemble sizes (integer)

See Also
--------
boss.pp.pp_main.PPMain : Main post-processing interface
boss.pp.plotter.Plotter : Creates diagnostic plots
boss.bo.results.BOResults : Container for optimization results
"""
from pathlib import Path

import numpy as np

from boss.bo.results import BOResults
from boss.pp.mesh import Mesh


class Dumper:
    """Data export interface for writing Bayesian optimization results to text files.

    The `Dumper` class provides methods for exporting various optimization data to
    human-readable text files. It uses `Mesh` objects to define coordinate grids for
    evaluating and dumping GP models, acquisition functions, and true functions over
    the input space.

    Parameters
    ----------
    results : BOResults
        Bayesian optimization results object containing models, data, and settings.
    mesh : Mesh
        Mesh object defining the coordinate grid for function evaluation.

    Attributes
    ----------
    results : BOResults
        The BO results object.
    settings : Settings
        Settings object from the BO results.
    bounds : NDArray
        Optimization bounds from settings.
    mesh : Mesh
        The mesh object for coordinate grids.
    grid_slice : ArrayLike or None
        Grid specification from the mesh, if available.
    """
    def __init__(self, results: BOResults, mesh: Mesh) -> None:
        self.results = results
        self.settings = results.settings
        self.bounds = self.settings["bounds"]
        self.mesh = mesh
        try:
            self.grid_slice = mesh.get_grid_spec()
        except RuntimeError:
            self.grid_slice = None

    @classmethod
    def from_grid_slice(cls, results: BOResults, grid_slice=None):
        """Create a Dumper from a grid specification.

        This is a convenience constructor that creates a `Mesh` object from a grid
        specification before initializing the `Dumper`.

        Parameters
        ----------
        results : BOResults
            Bayesian optimization results object.
        grid_slice : ArrayLike or None, optional
            Grid specification for the mesh. If None, uses the grid spec from
            `results.settings["pp_model_slice"]`.

        Returns
        -------
        Dumper
            Initialized dumper instance.
        """
        if grid_slice is None:
            grid_slice = results.settings["pp_model_slice"]
        mesh = Mesh.from_grid_spec(results.settings["bounds"], grid_slice)
        self = cls(results, mesh)
        return self

    def _build_grid_header(self, name: str) -> str:
        shape = self.mesh.grid_pts
        header = f"# {name}, grid of "
        if len(shape) > 1:
            header += f"{shape[0]}x{shape[1]}="
        header += f"{np.prod(shape)} pts\n"
        return header

    def dump_columns(
        self,
        file_path: str | Path,
        col_data,
        header: str = "data",
        formats: str | dict[int, str] = "%18.9e",
    ) -> None:
        """Write column data to a text file with a header.

        This is the core utility method used by all other dump methods to write
        formatted numerical data to text files. It handles column stacking,
        format specification, and header writing.

        Parameters
        ----------
        file_path : str or Path
            Path to the output file.
        col_data : tuple or list of array_like
            Tuple or list of column data arrays to write. Each element can be
            1D (single column) or 2D (multiple columns).
        header : str, default="data"
            Header string to write at the top of the file. Should start with '#'.
        formats : str or dict[int, str], default="%18.9e"
            Output format specification. If str, applies to all columns. If dict,
            maps column indices to format strings. Non-specified columns use
            "%18.8e" as default.
        """
        num_cols = len(col_data)
        if isinstance(formats, str):
            fmt = formats
        elif isinstance(formats, dict):
            fmt = []
            for k in range(num_cols):
                if k in formats.keys():
                    fstr = formats[k]
                else:
                    fstr = "%18.8e"
                col = np.asarray(col_data[k])
                if len(col.shape) == 1:
                    reps = 1
                else:
                    reps = col.shape[1]
                fmt.extend([fstr] * reps)

        data_combined = np.column_stack(col_data)
        with open(file_path, "w") as fd:
            fd.write(header + "\n")
            np.savetxt(fd, data_combined, fmt=fmt)

    def dump_hyperparams(self, file_path: str | Path) -> None:
        """Write Gaussian process hyperparameter evolution to a text file.

        Exports the GP kernel hyperparameters (variance and lengthscales) for each
        iteration to a multi-column text file. Each row contains the iteration number,
        ensemble size, and all hyperparameter values.

        Parameters
        ----------
        file_path : str or Path
            Path to the output file (typically "hyperparameters.dat").

        Notes
        -----
        Output file columns: iteration, ensemble_size, variance, lengthscale_1, ..., lengthscale_d
        where d is the dimensionality of the optimization problem.
        """
        res = self.results
        header = "Model hyperparameter values by iteration (iter npts parameters)"
        iters = np.array(list(res["model_params"].keys()))
        npts = res.batch_tracker.ensemble_sizes[iters]
        model_params = res["model_params"].to_array()
        formats = {0: "%d", 1: "%5d"}
        self.dump_columns(
            file_path, (iters, npts, model_params), header, formats=formats
        )

    def dump_model(self, file_path: str | Path, itr: int=-1) -> None:
        """Write Gaussian process model predictions to a text file.

        Evaluates the GP model (mean and standard deviation) over the mesh grid
        and writes the results to a multi-column text file. For problems with more
        than 2 active dimensions, inactive dimensions are fixed to the predicted
        minimum location.

        Parameters
        ----------
        file_path : str or Path
            Path to the output file (typically in "data_models/" subdirectory).
        itr : int, default=-1
            Iteration number for which to dump the model. Negative indices count
            from the end (e.g., -1 is the last iteration).

        Notes
        -----
        Output file columns: x1, [x2, ...], mu, nu
        where x1, x2, ... are coordinate values, mu is the predicted mean, and
        nu is the predicted standard deviation.

        The header also includes the model's hyperparameter values at the
        specified iteration.
        """
        self.mesh.fix_dim_preset(self.results, preset="min", itr=itr)
        mesher = self.mesh
        model = self.results.reconstruct_model(itr)
        model_params = model.get_all_params()

        header = self._build_grid_header("Model output (x mu nu)")
        header += (
            "# Model parameter values "
            + "({}): ".format(" ".join(model_params.keys()))
            + format(tuple(model_params.values()))
        )
        Y, Y_std = mesher.evaluate_func(model.predict, coord_shaped=True)
        self.dump_columns(file_path, (mesher.coords, Y, Y_std), header)

    def dump_acq_func(self, file_path: str | Path, itr: int=-1) -> None:
        """Write acquisition function values to a text file.

        Evaluates the acquisition function over the mesh grid and writes the results
        to a multi-column text file. For problems with more than 2 active dimensions,
        inactive dimensions are fixed to the next acquisition point location.

        Parameters
        ----------
        file_path : str or Path
            Path to the output file (typically in "data_acqfns/" subdirectory).
        itr : int, default=-1
            Iteration number for which to dump the acquisition function. Negative
            indices count from the end (e.g., -1 is the last iteration).

        Notes
        -----
        Output file columns: x1, [x2, ...], alpha
        where x1, x2, ... are coordinate values and alpha is the acquisition
        function value.
        """
        mesh = self.mesh
        mesh.fix_dim_preset(self.results, preset='next_acq', itr=itr)
        header = self._build_grid_header("# Acquisition function (x af)")
        acqfn = self.results.reconstruct_acq_func(itr=itr)
        acq_vals = mesh.evaluate_func(acqfn, coord_shaped=True)[0]
        self.dump_columns(file_path, (mesh.coords, acq_vals), header)

    def dump_true_func(self, file_path: str | Path) -> None:
        """Write true objective function values to a text file.

        Evaluates the true objective function over the mesh grid and writes the
        results to a multi-column text file. This is useful for comparing GP model
        predictions against ground truth. For problems with more than 2 active
        dimensions, inactive dimensions are fixed to the predicted minimum location
        from the final iteration.

        Parameters
        ----------
        file_path : str or Path
            Path to the output file (typically "true_func.dat").

        Notes
        -----
        Output file columns: x1, [x2, ...], f(x)
        where x1, x2, ... are coordinate values and f(x) is the true objective
        function value.
        """
        mesh = self.mesh
        mesh.fix_dim_preset(self.results, preset='min', itr=-1)
        header = self._build_grid_header("True function output (x f)")
        f_user = lambda X: self.settings.f.evaluate(X).Y
        output = self.mesh.evaluate_func(
            f_user, coord_shaped=True
        )
        self.dump_columns(file_path, (mesh.coords, output[0]), header)

    def dump_convergence(self, file_path: str | Path) -> None:
        """Write convergence diagnostics to a text file.

        Computes and exports convergence metrics that track changes in the predicted
        global minimum over iterations. These metrics help assess whether the
        optimization has converged.

        Parameters
        ----------
        file_path : str or Path
            Path to the output file (typically "convergence_measures.dat").

        Notes
        -----
        Output file columns:
        - iteration: Iteration number
        - npts: Ensemble size (number of training points)
        - dx_glmin: Euclidean norm of change in predicted minimum location
        - dmu_glmin_normalized: Absolute change in predicted minimum value,
          normalized by the estimated y-range

        The convergence metrics start from iteration 1 (not 0) since they measure
        changes between consecutive iterations.
        """
        header = (
            "# Convergence measures by iteration"
            " (iter npts dx_glmin abs(dmu_glmin)/yrange)\n"
        )
        results = self.results
        iters = list(results["x_glmin"].keys())[1:]
        npts = results.batch_tracker.ensemble_sizes[iters]
        x_glmin = results["x_glmin"].to_array()
        mu_glmin = results["mu_glmin"].to_array()
        dmu_glmin = np.abs(np.diff(mu_glmin))
        dy = np.zeros_like(dmu_glmin)
        for ind, itr in enumerate(iters):
            y_min, y_max = results.get_est_yrange(itr)
            dy[ind] = y_max - y_min
        dx_glmin = np.linalg.norm(np.diff(x_glmin, axis=0), axis=1)
        self.dump_columns(
            file_path,
            (iters, npts, dx_glmin, dmu_glmin / dy),
            header,
            formats={0: "%d", 1: "%5d"},
        )

    def dump_min_predictions(self, file_path: str | Path) -> None:
        """Write predicted global minimum locations and values to a text file.

        Exports the predicted global minimum (location, mean, and uncertainty) for
        each iteration. This provides a complete record of how the optimization's
        best estimate evolved over time.

        Parameters
        ----------
        file_path : str or Path
            Path to the output file (typically "minimum_predictions.dat").

        Notes
        -----
        Output file columns:
        - iteration: Iteration number
        - npts: Ensemble size (number of training points)
        - x_glmin: Predicted minimum location (d columns for d-dimensional problems)
        - mu_glmin: GP mean at the predicted minimum
        - nu_glmin: GP standard deviation at the predicted minimum
        """
        results = self.results
        header = (
            "# Global minimum predictions by iteration"
            " (iter npts x_glmin mu_glmin nu_glmin)\n"
        )
        iters = list(results["x_glmin"].keys())
        npts = results.batch_tracker.ensemble_sizes[iters]
        x_glmin = results["x_glmin"].to_array()
        mu_glmin = results["mu_glmin"].to_array()
        nu_glmin = results["nu_glmin"].to_array()
        self.dump_columns(
            file_path,
            (iters, npts, x_glmin, mu_glmin, nu_glmin),
            header,
            formats={0: "%d", 1: "%5d"},
        )

    def dump_true_func_at_minima(self, file_path: str | Path) -> None:
        """Write true function values at predicted minima to a text file.

        Evaluates the true objective function at each iteration's predicted minimum
        location and compares it to the GP model's prediction. This provides insight
        into the quality of the GP model and the accuracy of minimum predictions.

        Parameters
        ----------
        file_path : str or Path
            Path to the output file (typically "true_f_at_x_hat.dat").

        Notes
        -----
        Output file columns:
        - iteration: Iteration number
        - npts: Ensemble size (number of training points)
        - f(x_glmin): True function value at the predicted minimum
        - error: Prediction error (f(x_glmin) - mu_glmin)
        """
        results = self.results
        header = (
            "# True function value at x_glmin locations by iteration"
            " (iter npts f(x_glmin) f(x_glmin)-mu_glmin)\n"
        )
        iters = list(results["x_glmin"].keys())
        npts = results.batch_tracker.ensemble_sizes[iters]
        x_glmin = results["x_glmin"].to_array()
        mu_glmin = results["mu_glmin"].to_array()
        f_true_hats = self.settings.f.evaluate(x_glmin).Y
        diff = f_true_hats - mu_glmin
        self.dump_columns(
            file_path,
            (iters, npts, f_true_hats, diff),
            header,
            formats={0: "%d", 1: "%5d"},
        )



"""Main post-processing interface for Bayesian optimization results.

This module provides the `PPMain` class, which serves as the high-level orchestrator
for post-processing BOSS optimization results. It coordinates the generation of
diagnostic plots and data dumps by delegating to specialized components:

- `boss.pp.plotter.Plotter`: Creates visualization plots (convergence, models,
  acquisition functions, hyperparameters)
- `boss.io.dumper.Dumper`: Writes numerical data to text files for external analysis
- `boss.pp.mesh.Mesh`: Manages coordinate grids for function evaluation

The `PPMain` class automates the entire post-processing workflow, creating a structured
directory (`postprocessing/`) containing:
- Hyperparameter evolution plots and data
- Convergence diagnostics
- Acquisition history visualization
- Per-iteration model and acquisition function outputs
- True function visualizations (when available)

Typical Usage
-------------
Post-processing is typically run after a BO optimization completes:

>>> from boss.bo.bo_main import BOMain
>>> from boss.pp.pp_main import PPMain
>>> bo = BOMain(func, bounds, kernel="rbf", initpts=5, iterpts=10)
>>> res = bo.run()
>>> pp = PPMain(res, pp_models=True, pp_acq_funcs=True)
>>> pp.run()

This creates the `postprocessing/` directory with all diagnostic outputs.

Alternatively, post-processing can be run from saved restart files:

>>> pp = PPMain.from_file("boss.rst", "boss.out")
>>> pp.run()

See Also
--------
boss.pp.plotter.Plotter : Creates diagnostic plots
boss.io.dumper.Dumper : Writes data files
boss.bo.bo_main.BOMain : Main optimization interface
"""
from __future__ import annotations

import shutil
from pathlib import Path
from typing import Any, Literal

import numpy as np

from boss.bo.results import BOResults
from boss.io.dumper import Dumper
from boss.io.main_output import MainOutput
from boss.pp.mesh import Mesh
from boss.pp.plotter import Plotter
from boss.settings import Settings
from copy import deepcopy
import copy


class PPMain:
    """Main post-processing orchestrator for Bayesian optimization results.

    The `PPMain` class manages the complete post-processing workflow, coordinating
    the `Plotter` and `Dumper` objects to generate diagnostic visualizations and
    data files. It creates a structured output directory with subdirectories for
    different types of outputs.

    Parameters
    ----------
    results : BOResults
        Bayesian optimization results object containing models, data, and settings.
    **pp_settings : Any
        Optional keyword arguments to override post-processing settings. Common options include:
        - pp_models : bool
            Whether to output model plots and data files (default: False)
        - pp_acq_funcs : bool
            Whether to output acquisition function plots and data (default: False)
        - pp_iters : array_like
            Specific iterations to post-process (default: all iterations)
        - pp_model_slice : array_like
            Grid specification for model/acquisition plots as [dim1, dim2, npts]
        - pp_truef_npts : int
            Number of grid points for true function visualization

    Attributes
    ----------
    results : BOResults
        The BO results object.
    settings : Settings
        Combined settings from results and pp_settings overrides.
    rstfile : str
        Path to the restart file.
    outfile : str
        Path to the output file.
    main_output : MainOutput
        Output handler for logging.
    basepath : Path
        Base directory for post-processing outputs ("postprocessing/").
    mesh : Mesh
        Mesh object for coordinate grid generation.
    plotter : Plotter
        Plotter object for creating visualizations.
    dumper : Dumper
        Dumper object for writing data files.

    See Also
    --------
    boss.pp.plotter.Plotter : Creates diagnostic plots
    boss.io.dumper.Dumper : Writes data files
    boss.bo.results.BOResults : Container for optimization results
    """
    def __init__(
        self,
        results: BOResults,
        **pp_settings: Any,
    ) -> None:
        keywords = results.settings
        keywords.update({k: v for k, v in pp_settings.items() if v is not None})
        settings = Settings(keywords.data, f=results.settings.f)
        settings.correct()

        self.results = results
        self.settings = settings
        self.rstfile = self.settings["rstfile"]
        self.outfile = self.settings["outfile"]
        self.main_output = None
        self._setup()

    @classmethod
    def from_file(
        cls, rstfile: str, outfile: str, main_output: MainOutput | None = None
    ) -> "PPMain":
        """Create a PPMain instance from saved restart and output files.

        This alternative constructor loads BO results and settings from disk,
        allowing post-processing to be performed independently from the optimization run.

        Parameters
        ----------
        rstfile : str
            Path to the restart file (typically "boss.rst") containing settings
            and restart data.
        outfile : str
            Path to the output file (typically "boss.out") containing iteration
            history and results.
        main_output : MainOutput or None, optional
            Optional output handler for logging. If None, a new one is created.

        Returns
        -------
        PPMain
            Initialized PPMain instance ready for post-processing.
        """
        self = cls.__new__(cls)
        self.results = BOResults.from_file(rstfile, outfile)
        self.settings = Settings.from_file(rstfile)
        self.rstfile = rstfile
        self.outfile = outfile
        self.main_output = main_output
        cls._setup(self)
        return self

    def _setup(self) -> None:
        """Initialize post-processing components and default settings.

        This internal method:
        - Sets default iteration range if not specified
        - Creates the MainOutput handler if needed
        - Initializes the Mesh object with grid specification
        - Creates Plotter and Dumper instances for output generation

        The method converts 1-indexed pp_model_slice dimensions to 0-indexed
        for internal use by the Mesh.
        """
        if self.settings["pp_iters"] is None:
            self.settings["pp_iters"] = np.arange(0, self.settings["iterpts"] + 1)

        if not getattr(self, "main_output", None):
            self.main_output = MainOutput(self.settings)
        self.basepath = Path("postprocessing")

        # Create objects to handle mesh, plots and dumps.
        self.zero_indexed_slice = copy.copy(self.settings["pp_model_slice"])
        self.zero_indexed_slice[0] -= 1
        self.zero_indexed_slice[1] -= 1
        self.mesh = Mesh.from_grid_spec(
            self.settings["bounds"], self.zero_indexed_slice
        )
        self.plotter = Plotter(self.results, self.mesh)
        self.dumper = Dumper(self.results, self.mesh)

    def run(self) -> None:
        """Execute the complete post-processing workflow.

        This method orchestrates the entire post-processing pipeline by:

        1. Creating the output directory structure (`postprocessing/` and subdirectories)
        2. Generating global diagnostic outputs:
           - Hyperparameter evolution plots and data files
           - Minimum predictions over iterations
           - Convergence diagnostics
           - Acquisition history visualization
        3. Processing per-iteration outputs (if enabled):
           - Model predictions (mean and variance) as plots and data files
           - Acquisition function values as plots and data files
        4. Generating true function outputs (if available and enabled):
           - True function values at predicted minima
           - True function visualization over the input space
        """
        if self.results is None:
            self.results = BOResults.from_file(self.rstfile, self.outfile)

        # Create the main pp directory.
        if self.basepath.is_dir():
            print("warning: overwriting directory 'postprocessing'")
        shutil.rmtree("postprocessing", ignore_errors=True)
        self.basepath.mkdir(exist_ok=True)
        self.create_subdirs()

        # process hyperparameters
        self.plotter.plot_hyperparameters(
            self.basepath / "hyperparameters.png", show=False
        )
        self.dumper.dump_hyperparams(self.basepath / "hyperparameters.dat")

        # process min preds
        self.dumper.dump_min_predictions(self.basepath / "minimum_predictions.dat")

        # process conv meas
        self.dumper.dump_convergence(self.basepath / "convergence_measures.dat")
        self.plotter.plot_convergence(
            save_as=self.basepath / "convergence_measures.png", show=False
        )

        # process acquisition locations
        self.plotter.plot_acquisitions(
            save_as=self.basepath / "acquisition_locations.png", show=False
        )

        # ensure minimum data is available for all iterations to be processed
        self.results.calc_missing_minima(self.settings["pp_iters"])

        # process per-iteration info
        for itr in self.settings["pp_iters"]:
            # acq. funcs
            if self.settings["pp_acq_funcs"]:
                file_dest = self._get_dest("acq", "data", itr)
                self.dumper.dump_acq_func(file_dest, itr=itr)
                file_dest = self._get_dest("acq", "graph", itr)
                self.plotter.plot_acq_func(itr=itr, save_as=file_dest, show=False)

            # models
            if self.settings["pp_models"]:
                file_dest = self._get_dest("model", "data", itr)
                self.dumper.dump_model(file_dest, itr=itr)
                file_dest = self._get_dest("model", "graph", itr)
                self.plotter.plot_model(itr=itr, save_as=file_dest, show=False)

        if self.settings["pp_truef_at_glmins"]:
            self.dumper.dump_true_func_at_minima(self.basepath / "true_f_at_x_hat.dat")
            self.plotter.plot_true_func_at_minima(
                self.basepath / "true_function_at_xhats.png", show=False
            )

        # process true function
        pp_truef_npts = self.settings["pp_truef_npts"]
        if pp_truef_npts:
            # Hack: temporarily change self.mesh to a new mesh with pp_truef_npts grid points in each dim
            mesh_modified = pp_truef_npts != self.mesh.grid_pts[0]
            if mesh_modified:
                pp_truef_slice = deepcopy(self.settings['pp_model_slice'])
                pp_truef_slice[0] -= 1
                pp_truef_slice[1] -= 1
                pp_truef_slice[2] = pp_truef_npts
                self.mesh = Mesh.from_grid_spec(
                    self.settings["bounds"], pp_truef_slice
                )

            self.plotter.plot_true_func(self.basepath / "true_func.png", show=False)
            self.dumper.dump_true_func(self.basepath / "true_func.dat")
            # revert back to the regular grid
            if mesh_modified:
                self.mesh = Mesh.from_grid_spec(
                    self.settings["bounds"], self.zero_indexed_slice
                )

    def _get_dest(
        self,
        target: Literal["model", "acq"],
        output_type: Literal["data", "graph"],
        itr: int,
    ) -> Path:
        """Generate output file path for per-iteration data or plots.

        Constructs the appropriate file path based on the target type (model or
        acquisition function), output type (data file or plot), and iteration number.
        File names follow the convention `it####_npts####.ext` where the first number
        is the iteration and the second is the ensemble size at that iteration.

        Parameters
        ----------
        target : {"model", "acq"}
            Type of output - "model" for GP model predictions, "acq" for acquisition function.
        output_type : {"data", "graph"}
            Format of output - "data" for text files (.dat), "graph" for plots (.png).
        itr : int
            Iteration number for which to generate the file path.

        Returns
        -------
        Path
            Full path to the output file.
        """
        templ = "it{:04}_npts{:04}"
        dests = {
            ("model", "data"): str(self.basepath) + "/data_models/{}.dat",
            ("model", "graph"): str(self.basepath) + "/graphs_models/{}.png",
            ("acq", "data"): str(self.basepath) + "/data_acqfns/{}.dat",
            ("acq", "graph"): str(self.basepath) + "/graphs_acqfns/{}.png",
        }
        npts = self.results.batch_tracker.ensemble_sizes[itr]
        file_dest = Path(dests[target, output_type].format(templ.format(itr, npts)))
        return file_dest

    def create_subdirs(self) -> None:
        """Create subdirectory structure for post-processing outputs.
        """
        if self.settings["pp_models"]:
            (self.basepath / "data_models").mkdir(exist_ok=True)
            (self.basepath / "graphs_models").mkdir(exist_ok=True)
        if self.settings["pp_acq_funcs"]:
            (self.basepath / "data_acqfns").mkdir(exist_ok=True)
            (self.basepath / "graphs_acqfns").mkdir(exist_ok=True)
        if self.settings["pp_local_minima"] is not None:
            (self.basepath / "data_local_minima").mkdir(exist_ok=True)



from __future__ import annotations

from abc import ABC, abstractmethod

from numpy.typing import ArrayLike, NDArray


class Model(ABC):
    """
    Base class for surrogate models used in Bayesian optimization.
    """

    @property
    @abstractmethod
    def dim(self) -> int:
        pass

    @property
    @abstractmethod
    def X(self) -> NDArray:
        pass

    @property
    @abstractmethod
    def Y(self) -> NDArray:
        pass

    @abstractmethod
    def add_data(self, X_new: ArrayLike, Y_new: ArrayLike) -> None:
        """
        Updates the model evidence (observations) dataset appending.
        """
        pass

    @abstractmethod
    def redefine_data(self, X: ArrayLike, Y: ArrayLike) -> None:
        """
        Updates the model evidence (observations) dataset overwriting.
        """
        pass

    @abstractmethod
    def get_best_xy(self) -> tuple[NDArray, float]:
        """
        Returns the lowest energy acquisition (x, y).
        """
        pass

    @abstractmethod
    def predict(
        self, x: ArrayLike, noise: bool = True, norm: bool = False
    ) -> tuple[NDArray, NDArray]:
        """
        Returns model prediction mean and variance at point x, with or without
        model variance (noise) and normalisation (norm).
        """
        pass

    @abstractmethod
    def predict_grads(
        self, x: ArrayLike, norm: bool = False
    ) -> tuple[NDArray, NDArray]:
        """
        Returns prediction mean and variance gradients with respect to input
        at point x, with or without normalisation (norm).
        """
        pass

    @abstractmethod
    def predict_mean_sd_grads(
        self, x: ArrayLike, noise: bool = True, norm: bool = True
    ) -> tuple[NDArray, NDArray, NDArray, NDArray]:
        """
        Returns the model prediction mean, standard deviation and their
        gradients at point x, with or without model variance (noise) and
        normalisation (norm).
        """
        pass

    @abstractmethod
    def predict_mean_grad(
        self, x: ArrayLike, norm: bool = True
    ) -> tuple[NDArray, NDArray]:
        """
        Returns model mean and its gradient at point x, with or without
        normalisation (norm).
        """
        pass

    def estimate_num_local_minima(self, search_bounds: ArrayLike) -> int:
        """
        Returns estimated number of local minima within bounds, calculated
        based on model properties.
        """
        raise NotImplementedError

    @abstractmethod
    def get_all_params(self, include_fixed=True) -> dict[str, NDArray]:
        """
        Returns model parameters as a dictionary.
        """
        pass

    @abstractmethod
    def get_unfixed_params(self) -> NDArray:
        """
        Returns the unfixed parameters of the model in an array.
        """
        pass

    @abstractmethod
    def sample_unfixed_params(self, num_samples: int):
        """
        Sample unfixed model parameters.
        """
        pass

    @abstractmethod
    def set_unfixed_params(self, params: NDArray) -> None:
        """
        Sets the unfixed parameters of the model to given values.
        """
        pass

    @abstractmethod
    def optimize(self, restarts: int = 1) -> None:
        """
        Updates unfixed model parameters.
        """
        pass



from __future__ import annotations

import abc
from collections import defaultdict
from dataclasses import dataclass
from pathlib import Path
from types import FunctionType
from typing import Any, Callable, Iterable, NamedTuple, Protocol, Type

import numpy as np
import sqlalchemy as sa
from numpy.typing import ArrayLike, NDArray
from sqlalchemy.orm import DeclarativeBase, Session, relationship, selectinload

import boss.keywords as bkw
from boss.bo.results import BOResults
from boss.settings import Settings


@dataclass
class ResultArrays:
    """
    Holds BOSS result arrays. This is the return type for Result records
    that are converted to numpy or aggregated.
    """

    X: NDArray | None = None
    Y: NDArray | None = None
    model_params: NDArray | None = None
    mu_glmin: NDArray | None = None
    nu_glmin: NDArray | None = None
    x_glmin: NDArray | None = None
    t_total: float | None = None


class Output(NamedTuple):
    """Class representing an output from a BOSS run."""

    name: str  # name of the output in BOResults
    sql_type: type  # type used for output in the db
    order: int = 0  # tensorial order of the output
    sparse: bool = False  # whether the output is a SparseList
    array: bool = False  # whether the output is a np.ndarray


# Add new outputs available from the BOResults object here
OUTPUTS = (
    Output(name="X", sql_type=sa.LargeBinary, order=2, array=True),
    Output(name="Y", sql_type=sa.LargeBinary, order=2, array=True),
    Output(name="x_glmin", sql_type=sa.LargeBinary, order=2, sparse=True),
    Output(name="mu_glmin", sql_type=sa.LargeBinary, order=1, sparse=True),
    Output(name="nu_glmin", sql_type=sa.LargeBinary, order=1, sparse=True),
    Output(name="model_params", sql_type=sa.LargeBinary, order=1, sparse=True),
    Output(name="t_total", sql_type=sa.Float, order=0),
)


class BaseKeywords(Protocol):
    """Protocol for an ORM model representing BOSS keywords.

    This is mainly used for documentation, type hinting and autocompletion.
    Concrete types inheriting this class are created dynamically at
    runtime by the create_keywords_type method.
    """

    @abc.abstractmethod
    def __init__(
        self,
        bo_settings: dict[str, Any] | Settings,
        **extra_column_values: Any,
    ) -> None:
        """
        Creates a new instance of a Keywords ORM model.

        Note that this instance represents a row/record in the
        keywords table of the benchmark database.

        Parameters
        ----------
        bo_settings : dict[str, Any] | Settings
            A dict or Settings object containing BOSS keywords that will
            be used as columns in the SQL table.
        extra_column_values : Any
            Values for extra, non-keyword columns that have been added
            as attributes to the class
        """
        pass

    @abc.abstractmethod
    def aggregate_results(self) -> ResultArrays:
        """Combines results from all BOSS run repetitions using the current keywords.

        For each Results record that is related to the Keywords,
        the result arrays are stacked together for easy access.

        Returns
        -------
        ResultArrays
            Stacked arrays from all BOSS runs. Example: We run 3 repetitions
            with the current keywords, and each repetition consists of 5 bo iterations
            (including initial points), then ResultArrays.mu_glmin will be a 3 x 5
            array of predicted global mins for each repetition / iteration.

        """
        pass


class BaseResults(Protocol):
    """Protocol for an ORM model representing BOSS results.

    This is mainly used for documentation purposes and type hinting.
    Concrete types inheriting this class are created dynamically at
    runtime by the create_results_type method.
    """

    @abc.abstractmethod
    def __init__(self, bo_results: BOResults) -> None:
        """
        Creates a new instance of the Results ORM model.

        Note that this instance represents a row/record in the
        results table of the benchmark database.

        Parameters
        ----------
        bo_results : BOResults
            BOResults
        """
        pass


def create_keywords_type(
    base: type[DeclarativeBase],
    keywords: Iterable[str] | None = None,
    **extra_columns: sa.Column,
) -> type[BaseKeywords]:
    """
    Creates a sqlalchemy ORM class that defines a table for BOSS keywords.

    The class is created dynamically (inheriting BaseKeywords)
    to provide flexibility in what BOSS keywords are included as columns.
    In a database, a Keywords record together one or more Result record (defined below)
    describes a single BOSS run.

    Parameters
    ----------
    base: type
        A type inherited from DeclarativeBase
    bo_keywords : Iterable[str] | None
        BOSS keywords that will be included as columns in the database table.
        If none are specified, ALL available keywords will be used.
    extra_columns : sqlalchemy.Column
        Extra columns to include in the table. Only supports types that don't
        require special initialization (e.g., int, float, str, bool).

    Returns
    -------
    Keywords : Dynamically created type 'Keywords'
        A sqlalchemy ORM model for the selected BOSS keywords.
    """
    attrs = {}
    attrs["__tablename__"] = "keywords"
    attrs["id"] = sa.Column(sa.Integer, primary_key=True)
    attrs["results"] = relationship("Results", back_populates="keywords")

    cat_map = {
        (bool, 0): sa.Boolean,
        (int, 0): sa.Integer,
        (str, 0): sa.String,
        (float, 0): sa.Float,
        (int, 1): sa.String,
        (int, 2): sa.String,
        (float, 1): sa.String,
        (float, 2): sa.String,
        (str, 1): sa.String,
        (str, 2): sa.String,
    }

    # If explicit BOSS keyword names to include in the database are given use only them.
    if keywords is not None:
        for key in keywords:
            cat = bkw.find_category(key)
            try:
                attrs[key] = sa.Column(cat_map[cat], nullable=True)
            except KeyError:
                raise KeyError(f"{key} is not a recognized BOSS keyword")
    # No explicit keyword names given, so we use all BOSS keywords.
    else:
        for cat in bkw.categories:
            for key in bkw.categories[cat]:
                attrs[key] = sa.Column(cat_map[cat], nullable=True)

    for name, col in extra_columns.items():
        attrs[name] = col

    def __init__(
        self,
        bo_settings: dict[str, Any] | Settings,
        **extra_column_values: Any,
    ) -> None:
        keyword_overlap = set(bo_settings.keys()) & bkw.get_keyword_names()
        for key in keyword_overlap:
            cat = bkw.find_category(key)
            try:
                if cat_map[cat] == sa.String:
                    setattr(self, key, bkw.stringify(bo_settings[key]))
                else:
                    setattr(self, key, bo_settings[key])

            except KeyError:
                raise KeyError(f"{key} is not a recognized BOSS keyword")

        for col_name, col_val in extra_column_values.items():
            setattr(self, col_name, col_val)

    def aggregate_results(self) -> ResultArrays:
        aggregates = defaultdict(list)
        exceptions = ["Y"]
        outputs_mod = [out for out in OUTPUTS if out.name not in exceptions]
        for rec in self.results:
            for out in outputs_mod:
                if out.name in rec.__dict__.keys():
                    val = getattr(rec, out.name)
                    if out.order == 2:
                        aggregates[out.name].append(
                            np.frombuffer(val, dtype=np.float64).reshape(-1, rec.dim)
                        )
                    elif out.order == 1:
                        aggregates[out.name].append(
                            np.frombuffer(val, dtype=np.float64)
                        )
                    elif out.order == 0:
                        aggregates[out.name].append(val)

            # TODO: handle gradients and multi-task
            if "Y" in rec.__dict__.keys():
                aggregates["Y"].append(np.frombuffer(rec.Y, dtype=np.float64)[:, None])

        res_arrs = ResultArrays()
        for col, arrs in aggregates.items():
            setattr(res_arrs, col, np.stack(arrs))

        return res_arrs

    attrs["aggregate_results"] = aggregate_results
    attrs["__init__"] = __init__
    keywords_type = type("Keywords", (base,), attrs)
    return keywords_type


def create_results_type(base: Type[DeclarativeBase]) -> Type[BaseResults]:
    """
    Creates an sqlalchemy ORM class that defines a table for BOSS results.

    In a database, a Results record together with a Keywords record
    describes a single BOSS run. Note that many Results records can correspond
    to a single Keywords record. Note that BOSS result arrays are stored as
    binary blobs in the database and need to converted to arrays after retrieval.

    Parameters
    ----------
    base: type
        A type inherited from DeclarativeBase.

    Returns
    -------
    Results : BaseResults
        A sqlalchemy ORM model for BOSS results.
    """
    attrs = {}
    attrs["__tablename__"] = "results"
    attrs["id"] = sa.Column(sa.Integer, primary_key=True)
    attrs["keywords_id"] = sa.Column(sa.Integer, sa.ForeignKey("keywords.id"))
    attrs["keywords"] = relationship("Keywords", back_populates="results")
    attrs["dim"] = sa.Column(sa.Integer)

    for out in OUTPUTS:
        attrs[out.name] = sa.Column(out.sql_type, nullable=True)

    def __init__(self, bo_results: BOResults) -> None:
        self.dim = len(bo_results.settings["bounds"])

        for out in OUTPUTS:
            if out.array:
                setattr(self, out.name, bo_results[out.name].tobytes())
            elif out.sparse:
                setattr(self, out.name, bo_results[out.name].to_array(True).tobytes())
            else:
                setattr(self, out.name, bo_results[out.name])

    attrs["__init__"] = __init__
    results_type = type("Results", (base,), attrs)
    return results_type


@dataclass
class ColumnSpec:
    py_type: type
    get_value: Callable
    nullable: bool = False

    def to_column(self):
        type_map = {
            "int": sa.Integer,
            "float": sa.Float,
            "bool": sa.Boolean,
            "str": sa.String,
        }
        return sa.Column(type_map[self.py_type.__name__], nullable=self.nullable)


def get_callable_name(f: Callable) -> str:
    if isinstance(f, FunctionType):
        name = f.__name__
    else:
        name = getattr(f, "name", None)
        if name is None:
            name = f.__class__.__name__
    return name


def get_userfn_name(bo_settings: Settings | dict[str, Any]) -> str:
    if isinstance(bo_settings, Settings):
        f = bo_settings.f
    else:
        f = bo_settings["f"]
    return get_callable_name(f)


def get_existing_types(base: type) -> dict[str, type]:
    subclasses = base.__subclasses__()
    existing_types = {t.__name__: t for t in subclasses}
    return existing_types


class BenchmarkDatabase:
    """
    An interface to an SQL database that stores BOSS benchmark runs.

    This class simplifies querying and adding new records to the database
    so that users don't need any knowledge of sqlalchemy.
    """

    def __init__(
        self,
        name: str | Path,
        keyword_columns: Iterable[str] | None = None,
        **custom_columns: ColumnSpec,
    ) -> None:
        """
        Creates a new BenchmarkDatabase with selected BOSS keywords and other
        user-specified columns (optional).

        Parameters
        ----------
        name : str | Path
            Name of the database (including file extension).
        keywords : Iterable[str] | None
            BOSS keywords that
            will be included as columns in the database. If none are specified,
            ALL available keywords will be used.
        **custom_columns : ColumnSpec
            Additional kwargs describing custom columns to include in the database.
            Kwarg keys will be used as column names, values are ColumnSpec instances.

                Example: mycol = ColumnSpec(int, get_value) will create an integer
                column called mycol whose value is set using the get_value function.
                This function must be defined by the user and is called with Settings
                and row_id as argument.
        """
        # we always add a custom column for the objective function name
        self.name = Path(name)
        self.engine = sa.create_engine(f"sqlite:///{self.name}")
        f_spec = ColumnSpec(
            py_type=str, get_value=lambda sts, _: get_userfn_name(sts), nullable=False
        )
        self._col_specs = {"f": f_spec}
        self._col_specs.update(custom_columns)
        self._keyword_cols = set(keyword_columns) if keyword_columns else set()
        extra_cols = {n: cs.to_column() for n, cs in self._col_specs.items()}
        self._init_common(extra_cols)

    @classmethod
    def from_file(cls, name: str | Path, **custom_columns: ColumnSpec) -> BenchmarkDatabase:
        self = cls.__new__(cls)
        self.name = Path(name)
        self.engine = sa.create_engine(f"sqlite:///{self.name}")
        if not self.name.is_file():
            raise FileNotFoundError(f"Database {name} not found.")

        # add a custom column for the name of the objective function
        f_spec = ColumnSpec(
            py_type=str, get_value=lambda x, y: get_userfn_name(x), nullable=False
        )
        self._col_specs = {"f": f_spec}
        self._col_specs.update(custom_columns)

        inspector = sa.inspect(self.engine)
        col_info = inspector.get_columns("keywords")
        col_names = set([col["name"] for col in col_info])

        # extract the names of columns that correspond to BOSS keywords
        bkw_names = bkw.get_keyword_names()
        kw_col_names = col_names & bkw_names
        self._keyword_cols = kw_col_names

        # extract remaining non-keyword and non-custom cols
        other_col_names = col_names - bkw_names - self._col_specs.keys() - {"id"}
        other_col_info = filter(lambda c: c["name"] in other_col_names, col_info)

        # create dict mapping names to sa.Column for all non-BOSS columns
        extra_cols = {n: cs.to_column() for n, cs in self._col_specs.items()}
        other_cols = {
            c["name"]: sa.Column(c["type"], nullable=c["nullable"])
            for c in other_col_info
        }
        extra_cols.update(other_cols)
        self._init_common(extra_cols)
        return self

    def _init_common(self, extra_columns: dict[str, sa.Column]) -> None:
        """Common initialization for all factory methods"""
        Base = type("Base", (DeclarativeBase,), {})
        Keywords = create_keywords_type(Base, self._keyword_cols, **extra_columns)
        self.Keywords = Keywords
        self.Results = create_results_type(Base)
        Base.metadata.create_all(bind=self.engine)
        self._id_last = 0  # id of the last inserted Keyword row

    def select(
        self, result_names: Iterable[str] | str | None = "all", **bo_keywords: Any
    ) -> list:
        """
        Selects a subset of keyword and result records from the database.

        Parameters
        ----------
        result_names : Iterable[str] | str | None
            Names of the results that should be returned. Defaults to
            all possible results: X, Y, model_params, x_glmin, mu_glmin, nu_glmin

        bo_keywords : Any
            Keyword values to filter the database by.

        Returns
        -------
        Records matching the selected keywords. Each records contains
        results (as specified by the result names) for all corresponding runs.
        """
        with Session(self.engine) as session:
            if result_names is None:
                opts = None
            elif result_names == "all":
                opts = selectinload(self.Keywords.results)
            else:
                col_names = ["dim"] + list(result_names)
                col_attrs = [getattr(self.Results, c) for c in col_names]
                opts = selectinload(self.Keywords.results).load_only(*col_attrs)
            if opts is None:
                output = session.execute(
                    sa.select(self.Keywords).filter_by(**bo_keywords)
                )
            else:
                output = session.execute(
                    sa.select(self.Keywords).options(opts).filter_by(**bo_keywords)
                )
            records = [t[0] for t in output]
        return records

    def add_keywords(self, bo_settings: Settings | dict[str, Any]) -> BaseKeywords:
        """
        Creates and adds a Keywords record to the database.

        Parameters
        ----------
        bo_settings : Settings | dict[str, Any]
            BOSS keywords to be included in the record.
        """
        extra_col_vals = {
            name: spec.get_value(bo_settings, self._id_last + 1)
            for name, spec in self._col_specs.items()
        }

        # If we only use a subset of BOSS keywords as columns we
        # filter those keywords here.
        if self._keyword_cols:
            settings = {k: v for k, v in bo_settings.items() if k in self._keyword_cols}
        else:
            settings = bo_settings

        keywords = self.Keywords(settings, **extra_col_vals)

        with Session(self.engine) as session:
            session.add(keywords)
            session.commit()
            self._id_last = keywords.id

        return keywords

    def add_results(
        self, bo_results: BOResults, keywords: BaseKeywords | None = None
    ) -> BaseResults:
        """
        Creates and adds a Results record to the database.

        Parameters
        ----------
        bo_results : BOResults
            BOSS results object to create the record from.
        keywords : BaseKeywords | None
            A Keywords record for the keywords used in the run.
            The new Results records will be related to this keyword
            record for ease of access.
        """
        results = self.Results(bo_results)
        if keywords is not None:
            results.keywords = keywords

        with Session(self.engine) as session:
            session.add(results)
            session.commit()
        return results
